/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string) gettext (string)

#define PLUGIN_NAME "bounce"

typedef struct _KibaBounce KibaBounce;
typedef struct _KibaBounceObject KibaBounceObject;

enum {
	KIBA_BOUNCE_UP = 1,
	KIBA_BOUNCE_DOWN = -1
};

struct _KibaBounceObject {
	KibaBounce *bounce;
	KibaObject *object;
	float offset;
	gint max_height;
	gint bounce_id;
	gint state;
	
	gboolean entered;
	gboolean urgent;
	
	gboolean scale_up;
	gboolean scaled;
	float scaleX;
	float scaleY;
	float scale_offset;
	gint scale_id;
};

typedef enum {
	KIBA_BOUNCE_URGENT = 0,
	KIBA_BOUNCE_SCALE,
	KIBA_BOUNCE_HOVERING,
	KIBA_BOUNCE_TRIGGERING
}KibaBounceFlags;

struct _KibaBounce {
	GList *objects;
	
	KibaBounceFlags flags;
	
	gint height;
	gint step;
};

/// scale and translate methods

static void place_surface (KibaObject *object,
				       float     *x,
				       float     *y,
				       gpointer    data) {
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*x = (d->halignment-1)*(bounce_object->offset + bounce_object->scale_offset);
		*y = 0.0;
	}
	else {
		*x = 0.0;
		*y = (d->valignment-1)*(bounce_object->offset + bounce_object->scale_offset);
	}
}

static void max_offset_func (KibaObject *object,
		float *left,
		float  *right,
		float  *bottom,
		float  *top,
		gpointer data)
{
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	
	dock = GTK_WIDGET(object)->parent;
	
	float max_offset = (float) bounce_object->bounce->height;
	
	switch(KIBA_DOCK(dock)->d.position) {
	case KIBA_POSITION_BOTTOM_EDGE:
		*left = 0.0f;
		*right = 0.0f;
		*bottom = 0.0f;
		*top = max_offset;
		break;
	case KIBA_POSITION_TOP_EDGE:
		*left = 0.0f;
		*right = 0.0f;
		*bottom = max_offset;
		*top = 0.0f;
		break;
	case KIBA_POSITION_LEFT_EDGE:
		*left = 0.0f;
		*right = max_offset;
		*bottom = 0.0f;
		*top = 0.0f;
		break;
	case KIBA_POSITION_RIGHT_EDGE:
		*left = max_offset;
		*right = 0.0f;
		*bottom = 0.0f;
		*top = 0.0f;
		break;
	default:
		g_assert_not_reached();
	}
}

static void scale_func (KibaObject *object,
				    float     *scale_x,
				    float     *scale_y,
				    gpointer    data) {
	KibaBounceObject *bounce_object = data;
	
	*scale_x = bounce_object->scaleX;
	*scale_y = bounce_object->scaleY;
}

static void max_scale_func (KibaObject *object,
				        float     *scale_x,
				        float     *scale_y,
				        gpointer    data) {
	KibaBounceObject *bounce_object = data;
	GtkWidget *dock;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		*scale_x = 1.0;
		*scale_y = 1.25;
	}
	else {
		*scale_x = 1.25;
		*scale_y = 1.0;
	}
}

/// accessing objects

static KibaBounceObject*
find_object (KibaBounce *bounce,
		KibaObject *object)
{
	GList *l;
	
	for (l = bounce->objects; l != NULL; l = l->next) {
		KibaBounceObject *bobject = l->data;
		if (bobject->object == object)
			return bobject;
	}
	
	return NULL;
}

static void
remove_object (KibaBounce *bounce,
		KibaObject *object)
{
	KibaBounceObject *bobject;
	
	bobject = find_object(bounce, object);
	if (bobject) {
		// bounce object active, clean it up
		
		if (bobject->bounce_id)
			g_source_remove (bobject->bounce_id);
		if (bobject->scale_id)
			g_source_remove (bobject->scale_id);
		
		kiba_object_remove_offset_function_by_data (
			    object, bobject);
		kiba_object_remove_scale_function_by_data (
			    object, bobject);
		
		bounce->objects = g_list_remove (bounce->objects, bobject);
		
		g_free (bobject);
	}
}

static KibaBounceObject*
add_object (KibaBounce *bounce,
		KibaObject *object)
{
	KibaBounceObject *bobject;
	
	bobject = g_new0 (KibaBounceObject, 1);
	bobject->object = object;
	bobject->offset = 0.0;
	bobject->max_height = 0;
	bobject->entered = FALSE;
	bobject->bounce = bounce;
	bobject->state = 0;
	bobject->bounce_id = 0;
	bobject->scaleX = 1.0;
	bobject->scaleY = 1.0;
	bobject->scale_offset = 0.0;
	bobject->urgent = FALSE;
	
	bounce->objects = g_list_append (bounce->objects, bobject);
	
	kiba_object_add_offset_function (object,
		place_surface,
		max_offset_func,
		FALSE,
		bobject);
	kiba_object_add_scale_function (object,
		scale_func,
		max_scale_func,
		FALSE,
		bobject);
	
	return bobject;
}

/// timeouts

static gint scale_object_timeout (gpointer data) {
	KibaBounceObject *object = data;
	GtkWidget *dock;
	float factor_step;
	const KibaLayoutData *d;
	
	dock = GTK_WIDGET(object->object)->parent;
	d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];
	
	float slow_down_factor = ((float) object->bounce->step)/10.0;
	factor_step = slow_down_factor*(0.04 + (0.04*(((float) object->bounce->step)/50.0)))*(object->scaleX/1.25);
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		if (object->scale_up) {
			object->scaleX -= factor_step;
			object->scaleY += factor_step;
			
			if (object->scaleY >= 1.5) {
				object->scaleY = 1.5;
				object->scaleX = 0.5;
				object->scale_up = FALSE;
				object->state = KIBA_BOUNCE_UP;
			}
		}
		else {
			object->scaleX += factor_step;
			object->scaleY -= factor_step;
			
			if (object->scaleY <= 1.0) {
				object->scaleY = 1.0;
				object->scaleX = 1.0;
				object->scale_up = TRUE;
			}
		}
		
		object->scale_offset = (GTK_WIDGET(object->object)->allocation.width*object->scaleX)*(object->scaleX-1.0f)/2.0f;
	}
	else {
		if (object->scale_up) {
			object->scaleY -= factor_step;
			object->scaleX += factor_step;
			
			if (object->scaleX >= 1.5) {
				object->scaleX = 1.5;
				object->scaleY = 0.5;
				object->scale_up = FALSE;
				object->state = KIBA_BOUNCE_UP;
			}
		}
		else {
			object->scaleY += factor_step;
			object->scaleX -= factor_step;
			
			if (object->scaleX <= 1.0) {
				object->scaleX = 1.0;
				object->scaleY = 1.0;
				object->scale_up = TRUE;
			}
		}
		
		object->scale_offset = (GTK_WIDGET(object->object)->allocation.height*object->scaleY)*(object->scaleY-1.0f)/2.0f;
	}
	
	kiba_object_update_offset (object->object);
	kiba_object_update_scale_factor (object->object);
	
	if (object->scaleX == 1.0) {
		object->scale_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static gboolean
continue_bouncing(KibaBounceObject *object)
{
	return (GET_FLAG(object->bounce->flags,
	    KIBA_BOUNCE_HOVERING) && object->entered) ||
	   (GET_FLAG(object->bounce->flags,
	    KIBA_BOUNCE_URGENT) && object->urgent);
}

static gint
bounce_object_timeout (gpointer data)
{
	KibaBounceObject *object = data;
	
	if (object->scale_id) {
		// wait for scale timeout
		return TRUE;
	}
	if (object->scaled) {
		object->scaled = FALSE;
		if (!continue_bouncing(object)) {
			// done with bouncing
			remove_object(object->bounce, object->object);
			return FALSE;
		}
	}
	
	// recalc offset
	float slow_down_factor = ((float) object->bounce->step)/10.0;
	object->offset +=  object->state *
		(slow_down_factor *(1.5 - object->offset/object->max_height));
	object->offset = MAX(0.0f, MIN (object->offset, (float) object->max_height));
	kiba_object_update_offset (object->object);
		
	if (object->offset == (float) object->max_height) {
		object->state = KIBA_BOUNCE_DOWN;
	}
	
	if (object->state == KIBA_BOUNCE_DOWN && object->offset <= 0.0) {
		gboolean stop_bounce = continue_bouncing(object);
		
		if (GET_FLAG(object->bounce->flags, KIBA_BOUNCE_SCALE)) {
			object->scale_up = TRUE;
			object->scale_id = g_timeout_add (20, scale_object_timeout, object);
			object->scaled = TRUE;
			return TRUE;
		} else {
			object->state = KIBA_BOUNCE_UP;
			object->scaled = TRUE;
		}
		
		if (stop_bounce || !object->max_height ||
		    GET_FLAG(object->object->state, KIBA_OBJECT_DRAGGING)) {
			// done with bouncing
			remove_object(object->bounce, object->object);
			return FALSE;
		}
	}
	
	return TRUE;
}

static void
start_bouncing (KibaBounceObject *object,
		gint bounce_height)
{
	if (bounce_height <= 0.0) {
		return;
	}
	if (GET_FLAG(object->object->state, KIBA_OBJECT_DRAGGING)) {
		return;
	}
	
	object->max_height = bounce_height;
	
	if (!object->bounce_id) {
		object->state = KIBA_BOUNCE_UP;
		object->bounce_id = g_timeout_add (20, bounce_object_timeout, object);
	}
}

/// preference callbacks

static void
kiba_prefs_bounce_height_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	bounce->height = GPOINTER_TO_INT(value);
}

static void
hovering_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	gboolean hovering = GPOINTER_TO_INT(value);
	
	if (hovering) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING)) {
			SET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING);
		}
	} else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING)) {
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING);
	}
}

static void
toggle_scale_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	gboolean scale = GPOINTER_TO_INT(value);
	
	if (scale) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE)) {
			SET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE);
		}
	} else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_SCALE)) {
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_SCALE);
	}
}

static void
toggle_urgent_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	gboolean urgent = GPOINTER_TO_INT(value);
	
	if (urgent) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT)) {
			SET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT);
		}
	} else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT)) {
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_URGENT);
	}
}

static void
toggle_triggering_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	gboolean triggering = GPOINTER_TO_INT(value);
	
	if (triggering) {
		if (!GET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING)) {
			SET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING);
		}
	} else if (GET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING)) {
		DEL_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING);
	}
}

static void
step_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaBounce *bounce = data;
	bounce->step = GPOINTER_TO_INT(value);
}

/// Widget events

static void
object_urgent_event_cb (KibaObject *object,
		gboolean urgent,
		gpointer data)
{
	KibaBounce *bounce = data;
	KibaBounceObject *bobject;
	
	bobject = find_object (bounce, object);
	if(urgent) {
		if (bobject) {
			bobject->urgent = TRUE;
		}
		if (GET_FLAG(bounce->flags, KIBA_BOUNCE_URGENT)) {
			if (!bobject) {
				bobject = add_object(bounce, object);
				bobject->urgent = TRUE;
			}
			start_bouncing(bobject, bounce->height);
		}
	} else {
		if (bobject) {
			bobject->urgent = FALSE;
		}
	}
}

static gboolean
object_release_event_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaBounce *bounce = data;
	KibaBounceObject *bobject;
	
	if (GET_FLAG(bounce->flags, KIBA_BOUNCE_TRIGGERING)) {
		bobject = find_object (bounce, KIBA_OBJECT(widget));
		if (!bobject) {
			bobject = add_object(bounce, KIBA_OBJECT(widget));
		}
		start_bouncing(bobject, bounce->height);
	}
	
	return FALSE;
}

static gboolean
object_enter_event_cb (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaBounce *bounce = data;
	KibaBounceObject *bobject = data;
	
	bobject = find_object (bounce, KIBA_OBJECT(widget));
	if (bobject) {
		bobject->entered = TRUE;
	}
	if (GET_FLAG(bounce->flags, KIBA_BOUNCE_HOVERING)) {
		if (!bobject) {
			bobject = add_object(bounce, KIBA_OBJECT(widget));
			bobject->entered = TRUE;
		}
		start_bouncing(bobject, bounce->height);
	}
	
	return FALSE;
}

static gboolean
object_leave_event_cb (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaBounce *bounce = data;
	KibaBounceObject *bobject = data;
	
	bobject = find_object (bounce, KIBA_OBJECT(widget));
	if (bobject) {
		bobject->entered = FALSE;
	}
	
	return FALSE;
}

static void
dock_object_add_cb (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	KibaBounce *bounce = data;
	
	if (!KIBA_IS_DRAWABLE_OBJECT(widget)) {
		return;
	}
	
	// connect to object signals
	g_signal_connect (G_OBJECT(widget), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), bounce);
	g_signal_connect (G_OBJECT(widget), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), bounce);
	g_signal_connect (G_OBJECT(widget), "button-release-event",
		G_CALLBACK(object_release_event_cb), bounce);
	g_signal_connect (G_OBJECT(widget), "urgent-toggled",
		G_CALLBACK(object_urgent_event_cb), bounce);
}

static void dock_object_remove_cb (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	KibaBounce *bounce = data;
	KibaBounceObject *bobject;
	
	// disconnect from object signals
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_enter_event_cb), bounce);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_leave_event_cb), bounce);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_release_event_cb), bounce);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_urgent_event_cb), bounce);
	
	remove_object(bounce, KIBA_OBJECT(widget));
}

/**
Plugin Functions
**/

static void plugin_unload (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaBounce *bounce;
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	bounce = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_add_cb), bounce);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_object_remove_cb), bounce);
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		dock_object_remove_cb (
			GTK_CONTAINER(dock),
			child->widget,
			bounce);
	}
	g_list_free (bounce->objects);
	
	g_free (bounce);
}

static gboolean plugin_load (KibaPlugin *plugin) {
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaBounce *bounce;
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	file = kiba_get_plugin_config_file(kiba, plugin);
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	bounce = g_new0(KibaBounce, 1);
	bounce->height = 20;
	bounce->step = 5;
	bounce->objects = NULL;
	bounce->flags = 0;
	
	kiba_prefs_add_callback (file, "bounce_height",
		kiba_prefs_bounce_height_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_speed",
		step_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_hovering",
		hovering_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_triggering",
		toggle_triggering_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_scale",
		toggle_scale_cb, bounce, TRUE);
	kiba_prefs_add_callback (file, "bounce_urgent",
		toggle_urgent_cb, bounce, TRUE);
	
	g_signal_connect (G_OBJECT(dock), "remove",
		G_CALLBACK(dock_object_remove_cb), bounce);
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(dock_object_add_cb), bounce);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		dock_object_add_cb (container, child, bounce);
	}
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) bounce);
	
	return TRUE;
}

static PrefMember bounce_options[] = {
    { "bounce_height", "0", KIBA_INT },
    { "bounce_triggering", "TRUE", KIBA_BOOLEAN },
    { "bounce_hovering", "TRUE", KIBA_BOOLEAN },
    { "bounce_scale", "TRUE", KIBA_BOOLEAN },
    { "bounce_urgent", "TRUE", KIBA_BOOLEAN },
    { "bounce_speed", "5", KIBA_INT }
};

static KibaPluginVtable vtable = {
	"Bounce Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
	
	// plugin version
	MAJOR_VERSION,
	MINOR_VERSION
};

KIBA_INIT_PLUGIN (vtable, bounce_options, NULL, NULL, NULL);
