/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <kiba.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-object.h>
#include <kiba-dock.h>
#include <kiba-screen.h>
#include <kiba-debug.h>
#include <kiba-desktop-icon.h>

#include "launcher.h"

// preference keys
#define PREFERENCES_ALPHA_KEY "launcher_alpha"
#define PREFERENCES_FACTOR_KEY "launcher_size_factor"
#define PREFERENCES_GROUPING_KEY "launcher_grouping"

#define APPLICATION_DIRECTORY "/usr/share/applications"

// some default applications added if launcher directory
// was not created yet
static const gchar* default_applications[] = {
	"firefox",
	"gimp",
	"pidgin",
	"gnome-terminal",
	"inkscape",
	"gedit",
	"transmission"
};

// preferences and default values
static PrefMember launcher_options[] = {
  { PREFERENCES_ALPHA_KEY, "1.0", KIBA_DOUBLE },
  { PREFERENCES_FACTOR_KEY, "1.0", KIBA_DOUBLE },
  { PREFERENCES_GROUPING_KEY, "FALSE", KIBA_BOOLEAN }
};

// helper

static void
add_default_applications(const gchar *launcher_directory)
{
	GDir *app_dir;
	GError *error = NULL;
	gchar *file_name;
	gint i;
	
	app_dir = g_dir_open (APPLICATION_DIRECTORY, 0, &error);
	if (error != NULL) {
		g_error_free (error);
		return;
	}

	while ((file_name = (gchar*) g_dir_read_name (app_dir)) != NULL) {
		if (!g_str_has_suffix (file_name, ".desktop")) {
			continue;
		}
		
		for (i = 0; i < G_N_ELEMENTS(default_applications); i++) {
			const gchar *app_name = default_applications[i];
			
			if (g_str_has_prefix (file_name, app_name)) {
				gchar *from = g_strdup_printf("%s/%s",
					APPLICATION_DIRECTORY, file_name);
				gchar *to = g_strdup_printf("%s/%s",
					launcher_directory, file_name);
				kiba_utils_copy_file(from, to);
			}
		}
	}
	
	g_dir_close (app_dir);
}

// KibaPrefs Callbacks

static void object_size_factor_cb (const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	
	launcher->cfg_scale = g_ascii_strtod((gchar*) value, NULL);
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		group = l->data;
		
		g_object_set (G_OBJECT (group->object),
			    "scale_x", launcher->cfg_scale, 
			    "scale_y", launcher->cfg_scale, 
			    NULL);
	}
}

static void
launcher_object_alpha_cb (const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	
	launcher->cfg_alpha = MIN(1.0f, MAX(0.0f,
		g_ascii_strtod((gchar*) value, NULL)));
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		group = l->data;
		kiba_drawable_object_set_alpha (
			KIBA_DRAWABLE_OBJECT(group->object),
			launcher->cfg_alpha);
	}
}

static void
launcher_grouping_cb (const gchar *option_id,
		gpointer value,
		gpointer data)
{
	KibaLauncher *launcher = data;
	
	launcher->cfg_grouping = (gboolean) GPOINTER_TO_INT(value);
	if (launcher->cfg_grouping == TRUE) {
		kiba_launcher_group_icons (launcher);
	} else {
		kiba_launcher_ungroup_icons (launcher);
	}
}

/// Plugin Functions

static void
launcher_plugin_unload (KibaPlugin *plugin)
{
	Kiba *kiba = kiba_get();
	KibaLauncher *launcher;
	KibaPrefsFile *configFile;
	
	launcher = g_object_get_data (G_OBJECT(plugin), LAUNCHER_PLUGIN_NAME);
	configFile = kiba_get_plugin_config_file(kiba, plugin);
	// remove preference callbacks
	kiba_prefs_remove_callback_by_data (configFile, launcher);
	// destroy the launcher
	kiba_launcher_destroy (launcher);
	
	g_object_set_data (G_OBJECT(plugin),
		LAUNCHER_PLUGIN_NAME, NULL);
}

static gboolean
launcher_plugin_load (KibaPlugin *plugin)
{
	Kiba *kiba = kiba_get();
	KibaLauncher *launcher;
	GError *error = NULL;
	const gchar *homedir;
	KibaPrefsFile *configFile;
	gchar *file_name, *launcher_directory, *dock_name;
	KibaWin *win;
	KibaDock *dock;
	GDir *dir;
	
#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, DATADIR"/locale");
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	dock_name = kiba_dock_get_name(dock);
	g_return_if_fail(dock_name != NULL);
	
	homedir = g_get_home_dir();
	// directory for .desktop files
	launcher_directory = g_strdup_printf(
		"%s/.kiba-dock/%s/" LAUNCHER_DIRECTORY_NAME,
		homedir, dock_name);
	g_free(dock_name);
	
	if (!g_file_test (launcher_directory, G_FILE_TEST_IS_DIR)) {
		gint i;
		
		if (!kiba_utils_recursive_mkdir (launcher_directory)) {
			DEBUG_ERROR ("Failed to create %s!\n", launcher_directory);
			g_free(launcher_directory);
			return FALSE;
		}
		// populate with some default applications
		add_default_applications(launcher_directory);
	}
	
	launcher = kiba_launcher_new(G_TYPE_MODULE(plugin), launcher_directory);
	if (launcher == NULL) {
		DEBUG_ERROR ("Failed to create the KibaLauncher Instance!\n");
		g_free(launcher_directory);
		return FALSE;
	}
	
	g_object_set_data (G_OBJECT(plugin),
		LAUNCHER_PLUGIN_NAME, (gpointer) launcher);
	
	configFile = kiba_get_plugin_config_file(kiba, plugin);
	
	// add pref callbacks
	kiba_prefs_add_callback (configFile, PREFERENCES_GROUPING_KEY,
		launcher_grouping_cb, launcher, TRUE);
	kiba_prefs_add_callback (configFile, PREFERENCES_ALPHA_KEY,
		launcher_object_alpha_cb, launcher, TRUE);
	kiba_prefs_add_callback (configFile, PREFERENCES_FACTOR_KEY,
		object_size_factor_cb, launcher, TRUE);
	
	// add desktop files from homedir
	dir = g_dir_open (launcher_directory, 0, &error);
	if (error != NULL) {
		DEBUG_ERROR ("Failed to open launcher directory '%s': %s\n",
			launcher_directory, error->message);
		g_error_free (error);
		g_free(launcher_directory);
		return FALSE;
	}
	
	while ((file_name = (gchar*) g_dir_read_name (dir)) != NULL) {
		KibaDesktopIcon *icon;
		
		if (!g_str_has_suffix (file_name, ".desktop")) {
			continue;
		}
		
		gchar uri[strlen(launcher_directory) + strlen(file_name) + 1];
		sprintf (uri, "%s/%s", launcher_directory, file_name);
		
		icon = kiba_desktop_icon_new ();
		if (kiba_desktop_icon_set_file (icon, uri)) {
			kiba_launcher_add (launcher, icon);
		} else {
			g_object_unref(G_OBJECT(icon));
		}
	}
	g_dir_close (dir);
	g_free(launcher_directory);
	
	return TRUE;
}

static KibaPluginVtable vtable = {
	// summary
	"Launcher Plugin for Kiba-Dock",
	// description
	NULL,
	
	// author
	"Daniel Bessler <danielb@kiba-dock.org>",
	// homepage
	"http://www.kiba-dock.org",
	
	// loadable flag
	KIBA_PLUGIN_MULTIPLE_LOADABLE,
	
	// load
	launcher_plugin_load,
	// unload
	launcher_plugin_unload,
	
	// plugin version
	MAJOR_VERSION,
	MINOR_VERSION
};

KIBA_INIT_PLUGIN(vtable,
	launcher_options,
	NULL, NULL, NULL);
