
#include "config.h"

#include <gdk/gdkkeysyms.h>

#include <kiba-desktop-editor.h>
#include <kiba-app-chooser.h>
#include <string.h>

#include "launcher.h"

#define APPLICATIONS_DIRECTORY "/usr/share/applications"
#define DESKTOP_SUFFIX ".desktop"

#define BUTTON_ENTRY_KEY "kiba-entry"

/// signal handler

static void
desktop_editor_file_saved_cb (KibaDesktopEditor *editor,
		const gchar *file,
		gpointer data)
{
	KibaDesktopIcon *saved_icon = data;
	KibaLauncher *launcher;
	KibaLauncherIconGroup *group;
	
	launcher = g_object_get_data(G_OBJECT(saved_icon), ICON_LAUNCHER_KEY);
	
	kiba_desktop_icon_set_file (saved_icon, file);
	// select the icon
	group = kiba_launcher_find_icon_group (launcher, saved_icon);
	if (group) {
		kiba_launcher_select_icon(launcher, group, saved_icon);
	}
}

static void
app_activated (KibaAppChooser *chooser,
		const gchar *desktop,
		gpointer user_data)
{
	KibaLauncher *launcher = user_data;
	KibaLauncherIconGroup *group;
	KibaDesktopIcon *icon;
	gchar *content;
	gchar *file_name;
	FILE *file;
	
	file_name = g_strrstr (desktop, "/") + 1;
	
	char uri[strlen(launcher->launcher_directory) + strlen(file_name) + 1];
	sprintf (uri, "%s/%s", launcher->launcher_directory, file_name);
	
	file = fopen (uri, "aw");
	if (file == NULL) {
		return;
	}
		
	if (!g_file_get_contents (desktop, &content, NULL, NULL)) {
		fclose (file);
		g_remove (uri);
		return;
	}
	
	// write to file
	fprintf (file, "%s", content);
	g_free (content);
	fclose (file);
	
	// create the launcher icon
	icon = kiba_desktop_icon_new ();
	if (kiba_desktop_icon_set_file (icon, uri)) {
		kiba_launcher_add (launcher, icon);
	} else {
		g_object_unref (G_OBJECT(icon));
		DEBUG_ERROR("Cannot add launcher %s.\n", uri);
		return;
	}
	
	// select the icon
	group = kiba_launcher_find_icon_group (launcher, icon);
	if (group) {
		kiba_launcher_select_icon(launcher, group, icon);
	}
}

static gboolean
filename_selected (GtkWidget *button,
		GdkEventButton *event,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaDesktopIcon *icon;
	KibaDesktopEditor *desktop_editor;
	GtkWidget *entry;
	FILE *launcher_file;
	
	// TODO: let user know about the errors happening here
	
	entry = g_object_get_data (G_OBJECT(button), BUTTON_ENTRY_KEY);
	
	const gchar *filename = gtk_entry_get_text (GTK_ENTRY(entry));
	if (!filename || strlen(filename) == 0) {
		// invalid filename
		return TRUE;
	}
	
	gchar *desktop_file = g_strdup_printf ("%s/%s%s",
		launcher->launcher_directory, filename, 
		(g_str_has_suffix (filename, DESKTOP_SUFFIX) ? "" : DESKTOP_SUFFIX));
	if (g_file_test(desktop_file, G_FILE_TEST_EXISTS)) {
		// file with the same name allready added
		g_free(desktop_file);
		return TRUE;
	}
	
	launcher_file = fopen (desktop_file, "w");
	if (!launcher_file) {
		g_free(desktop_file);
		return TRUE;
	}
	
	fprintf (launcher_file,
		"[Desktop Entry]\n"
		"Encoding=UTF-8\n"
		"Version=0.9.2\n"
		"Name=%s\n"
		"Comment=\n"
		"Description=\n"
		"Exec=%s\n"
		"Icon=" GTK_STOCK_MISSING_IMAGE "\n"
		KIBA_KEY_FILE_DESKTOP_KEY_GROUP "=%s\n"
		KIBA_KEY_FILE_DESKTOP_KEY_CLASS "=" KIBA_KEY_FILE_DESKTOP_DEFAULT_CLASS "\n",
		filename, filename, filename);
	
	fclose (launcher_file);
	
	icon = kiba_desktop_icon_new ();
	if (!kiba_desktop_icon_set_file (icon, desktop_file)) {
		g_object_unref (G_OBJECT(icon));
		g_free(desktop_file);
		return TRUE;
	}
	
	kiba_launcher_add (launcher, icon);
	
	desktop_editor = kiba_desktop_editor_new (desktop_file);
	gtk_widget_show (desktop_editor->win);
	
	g_signal_connect (G_OBJECT(desktop_editor), "file-saved",
		G_CALLBACK(desktop_editor_file_saved_cb), icon);
	
	// destroy dialog
	GtkWidget *toplevel = button;
	while (toplevel->parent) { toplevel = toplevel->parent; }
	gtk_widget_destroy(toplevel);
	
	g_free(desktop_file);
	
	return FALSE;
}
			
static gboolean
key_pressed (GtkWidget *widget,
		GdkEventKey *event,
		gpointer user_data)
{
	if (event->keyval == GDK_Return) {
		filename_selected(widget, NULL, user_data);
	}
	return TRUE;
}

static gboolean
dialog_destroy_cb (GtkWidget *button,
		GdkEventButton *event,
		gpointer data)
{
	gtk_widget_destroy (GTK_WIDGET(data));
	
	return TRUE;
}

/// public methods

void
kiba_launcher_populate_dock_create (KibaDock *dock,
		gpointer data)
{
	GtkWidget *win;
	GtkWidget *box;
	GtkWidget *label;
	GtkWidget *entry;
	GtkWidget *button;
	GtkWidget *button_box;
	GdkPixbuf *pixbuf = NULL;
	
	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title (GTK_WINDOW(win), _("Launcher creation dialog"));
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(win), TRUE);
	gtk_window_set_position (GTK_WINDOW(win), GTK_WIN_POS_CENTER);
	if (kiba_load_icon ("kiba-dock", 16, &pixbuf)) {
		gtk_window_set_icon (GTK_WINDOW(win), pixbuf);
		g_object_unref (G_OBJECT(pixbuf));
	}
	
	box = gtk_vbox_new(FALSE, 5); {
		label = gtk_label_new(_("Please enter a launcher name."));
		gtk_box_pack_start (GTK_BOX(box), label, TRUE, TRUE, 0);
		gtk_widget_show (label);
		
		entry = gtk_entry_new ();
		gtk_box_pack_start (GTK_BOX(box), entry, TRUE, TRUE, 0);
		gtk_widget_show (entry);
		
		button_box = gtk_vbox_new(FALSE, 5); {
			button = gtk_button_new_from_stock (GTK_STOCK_OK);
			gtk_box_pack_end (GTK_BOX(button_box), button, FALSE, FALSE, 0);
			g_object_set_data (G_OBJECT(button), BUTTON_ENTRY_KEY, entry);
			gtk_widget_show (button);

			g_signal_connect (G_OBJECT(button), "key-press-event",
				G_CALLBACK(key_pressed), data);
			g_signal_connect (G_OBJECT(button), "button-release-event",
				G_CALLBACK(filename_selected), data);
			g_signal_connect_after (G_OBJECT(button), "button-release-event",
				G_CALLBACK(dialog_destroy_cb), win);
		}
		gtk_box_pack_start (GTK_BOX(box), button_box, TRUE, TRUE, 0);
		gtk_widget_show (button_box);
	}
	gtk_container_add (GTK_CONTAINER(win), box);
	gtk_widget_show (box);
	gtk_window_present (GTK_WINDOW(win));
}

void
kiba_launcher_populate_dock (KibaDock *dock,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaAppChooser *chooser;
	
	chooser = kiba_app_chooser_new(APPLICATIONS_DIRECTORY);
	gtk_widget_show (chooser->view_win->win);
	gdk_window_set_keep_above (chooser->view_win->win->window, TRUE);
	
	g_signal_connect (G_OBJECT(chooser), "app-activate",
		G_CALLBACK(app_activated), launcher);
}

void
kiba_launcher_edid_icon (GtkWidget *widget,
		gpointer  data)
{
	KibaDesktopIcon *icon = data;
	KibaDesktopEditor *desktop_editor;
	
	desktop_editor = kiba_desktop_editor_new (icon->entry.uri);
	gtk_widget_show (desktop_editor->win);
	
	g_signal_connect (G_OBJECT(desktop_editor), "file-saved",
		G_CALLBACK(desktop_editor_file_saved_cb), icon);
}

