/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <string.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-menu-items.h>

#include "launcher.h"

/// declarations

static gboolean
button_press_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data);
static gboolean
button_release_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data);
static gboolean
scroll_cb (GtkWidget *widget,
		GdkEventScroll *event,
		gpointer data);
static void
drag_received_cb (GtkWidget *widget,
		GdkDragContext *drag_context,
		gint x,
		gint y,
		GtkSelectionData *selection_data,
		guint info,
		guint time,
		gpointer user_data);

/// helper

static KibaDesktopIcon*
get_active_group_icon(KibaLauncherIconGroup *group)
{
	return group->icons ? (KibaDesktopIcon*) group->icons->data : NULL;
}

static KibaLauncherIconGroup*
find_object_group (KibaLauncher *launcher,
		KibaObject *object)
{
	GList *l;
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		
		if (group->object == object) {
			return group;
		}
	}
	
	return NULL;
}

static void
create_icon_group (KibaLauncher *launcher,
		KibaDesktopIcon *icon)
{
	KibaLauncherIconGroup *group;
	GtkWidget *widget;
	gchar *object_id;
	
	// create the icon group
	group = g_new0 (KibaLauncherIconGroup, 1);
	if (launcher->cfg_grouping) {
		group->id = g_strdup (icon->entry.group);
	} else {
		group->id = g_strdup (icon->entry.name);
	}
	group->count = 1;
	group->launcher = launcher;
	group->icons = g_list_append (NULL, icon);
	group->view_win = NULL;
	
	launcher->groups = g_list_append (launcher->groups, group);
	
	// create the dock object
	object_id = g_strdup_printf(LAUNCHER_PLUGIN_NAME "_%s", group->id);
	widget = kiba_icon_new_from_file(object_id,
		launcher->cfg_scale, launcher->cfg_scale,
		icon->entry.icon);
	g_free(object_id);
	
	// add object to the dock
	group->object = KIBA_OBJECT(widget);
	kiba_object_set_title (group->object,
		icon->entry.name);
	kiba_drawable_object_set_alpha (KIBA_DRAWABLE_OBJECT(widget),
		launcher->cfg_alpha);
	kiba_add_plugin_object (kiba_get(),
		launcher->plugin, KIBA_OBJECT(widget));
	gtk_widget_show (widget);
	
	// connect to object signals
	g_signal_connect_after (G_OBJECT (widget), "button-press-event",
			G_CALLBACK (button_press_cb), launcher);
	g_signal_connect_after (G_OBJECT (widget), "button-release-event",
			G_CALLBACK (button_release_cb), launcher);
	g_signal_connect (G_OBJECT (widget), "scroll-event",
			G_CALLBACK (scroll_cb), launcher);
	g_signal_connect (G_OBJECT (widget), "drag-data-received",
			G_CALLBACK (drag_received_cb), launcher);
}

static void
kiba_launcher_icon_file_set_cb (KibaDesktopIcon *icon,
		const gchar *uri,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GList *l;
	
	if (!uri) {
		return;
	}
	
	// check if icon was added before, in this case, just reload the file
	group = NULL;
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *g = l->data;
		GList *m;
		for (m = g->icons; m != NULL; m = m->next) {
			if (m->data == icon) {
				group = g;
				break;
			}
			if (group) {
				break;
			}
		}
	}
	if (group) {
		kiba_drawable_object_switch_surface(
			KIBA_DRAWABLE_OBJECT(group->object));
		kiba_icon_set_file(
			KIBA_ICON(group->object), icon->entry.icon);
		return;
	}
	
	if (launcher->cfg_grouping) {
		for (l = launcher->groups; l != NULL; l = l->next) {
			group = l->data;
			
			if (strcmp (group->id, icon->entry.group) == 0) {
				GList *m;
				
				if (g_list_find (group->icons, icon)) {
					return;
				}
				
				group->icons = g_list_append (group->icons, icon);
				group->count++;
				
				for (m = launcher->groups; m != NULL; m = m->next) {
					KibaLauncherIconGroup *check_group = m->data;
					
					if (group == check_group) {
						continue;
					}
					
					if (!g_list_find (check_group->icons, icon)) {
						continue;
					}
					
					check_group->icons = g_list_remove (check_group->icons, icon);
					check_group->count--;
				}
				
				return;
			}
		}
	}
	
	create_icon_group (launcher, icon);
}

/// Callbacks

static gboolean
kiba_launcher_select_icon_cb (GtkWidget *widget,
		gpointer data)
{
	KibaLauncherIconGroup *group = data;
	KibaDesktopIcon *icon;
	GList *l;
	
	icon = g_object_get_data (G_OBJECT(widget), "desktop-icon");
	
	// switch object icon
	kiba_launcher_select_icon(group->launcher, group, icon);
	
	return TRUE;
}

static gboolean
kiba_launcher_remove_cb (GtkWidget *widget,
		gpointer data)
{
	KibaLauncherIconGroup *group = data;
	KibaDesktopIcon *icon = g_object_get_data(G_OBJECT(widget), "desktop-icon");
	
	kiba_launcher_remove (group->launcher, icon);
	if (g_file_test(icon->entry.uri, G_FILE_TEST_IS_SYMLINK)) {
		g_unlink (icon->entry.uri);
	} else if (g_file_test(icon->entry.uri, G_FILE_TEST_IS_REGULAR)) {
		g_remove (icon->entry.uri);
	}
	g_object_unref (G_OBJECT(icon));
	
	return TRUE;
}

static gboolean
button_press_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	GtkWidget *menu;
	GtkWidget *item, *image;
	GList *l;
	
	if (event->button != 3) {
		return FALSE;
	}
	
	menu = gtk_menu_new ();
	
	group = find_object_group (launcher, KIBA_OBJECT(widget));
	for (l = group->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		GtkWidget *groupmenu;
		GdkPixbuf *pixbuf = NULL;
		
		kiba_load_icon (icon->entry.icon, 24,  &pixbuf);
		
		if (group->count > 1) {
			groupmenu = gtk_menu_new ();
			
			item = gtk_image_menu_item_new_with_label (icon->entry.name);
			if (pixbuf) {
				image = gtk_image_new_from_pixbuf (pixbuf);
				gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			}
			
			gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
			gtk_menu_item_set_submenu (GTK_MENU_ITEM (item), GTK_WIDGET(groupmenu));
			gtk_widget_show (item);
		} else {
			groupmenu = menu;
		}
		
		item = gtk_image_menu_item_new_with_label (_("Edit"));
		image = gtk_image_new_from_stock (GTK_STOCK_EDIT, GTK_ICON_SIZE_DIALOG);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_launcher_edid_icon), icon);
		gtk_widget_show (item);
		
		kiba_menu_add_separator (groupmenu);
		
		if (group->icons->data != icon) {
			item = gtk_image_menu_item_new_with_label (_("Activate"));
			g_object_set_data (G_OBJECT(item), "desktop-icon", icon);
			image = gtk_image_new_from_icon_name (GTK_STOCK_SELECT_ALL, GTK_ICON_SIZE_MENU);
			gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
			gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
			g_signal_connect (G_OBJECT (item), "activate",
				G_CALLBACK (kiba_launcher_select_icon_cb), group);
			gtk_widget_show (item);
		}
	
		item = gtk_image_menu_item_new_with_label (_("Launch"));
		image = gtk_image_new_from_icon_name (GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_menu_item_activate_desktop_file), icon->entry.uri);
		gtk_widget_show (item);
		
		item = gtk_image_menu_item_new_with_label (_("Remove"));
		g_object_set_data (G_OBJECT(item), "desktop-icon", icon);
		image = gtk_image_new_from_icon_name (GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
		gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (item), image);
		gtk_menu_shell_append (GTK_MENU_SHELL (groupmenu), item);
		g_signal_connect (G_OBJECT (item), "activate",
			G_CALLBACK (kiba_launcher_remove_cb), group);
		gtk_widget_show (item);
		
		if (pixbuf) {
			g_object_unref (pixbuf);
		}
	}
	
	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		event->button, event->time);
	
	return TRUE;
}

static gboolean
button_release_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaDesktopIcon *icon;
	KibaLauncherIconGroup *group;
	
	if (event->button == 3) {
		return TRUE;
	}
	
	group = find_object_group (launcher, KIBA_OBJECT(widget));
	icon = g_list_nth_data (group->icons, 0);
	
	if (event->button == 2 && group->launcher->cfg_grouping) {
		kiba_launcher_icon_view_toggle (group);
		return TRUE;
	}
	
	if (icon->entry.exec != NULL) {
		kiba_object_launch (group->object, icon->entry.exec);
	}
	
	return TRUE;
}

static gboolean
scroll_cb (GtkWidget *widget,
		GdkEventScroll *event,
		gpointer data)
{
	KibaLauncher *launcher = data;
	KibaLauncherIconGroup *group;
	KibaDesktopIcon *newActive;
	GList *l;
	
	group = find_object_group (launcher, KIBA_OBJECT(widget));
	
	if (g_list_length(group->icons) <= 1) {
		return TRUE;
	}
	
	// switch object icon
	kiba_utils_glist_scroll (&group->icons, event->direction);
	newActive = group->icons->data;
	kiba_drawable_object_switch_surface (KIBA_DRAWABLE_OBJECT(widget));
	kiba_icon_set_file(KIBA_ICON(group->object), 
		newActive->entry.icon);
	kiba_object_set_title (group->object,
		newActive->entry.name);
	
	if (group->view_win) {
		icon_view_launcher_icon_selected_cb(launcher,
			newActive, group);
	}
	
	return TRUE;
}

/// drag and drop

static KibaDesktopIcon*
kiba_launcher_add_icon_from_uri (KibaLauncher *launcher,
		KibaLauncherIconGroup *group,
		GdkDragAction action,
		const gchar *launcherdir,
		const gchar *uri)
{
	static const char *file_prefix = "file://";
	static const char desktop_entry_group[] = "Desktop Entry";
	
	KibaDesktopIcon *icon;
	FILE *file;
	char *filename = NULL;
	gboolean intern_dnd = FALSE;
	gboolean intern_success = FALSE;
	GList *l;
	
	if (g_strrstr(uri, "/")) {
		filename = g_strdup(g_strrstr(uri, "/")+1);
	} if (!filename) {
		filename = g_strdup(uri);
	}
	
	char path[strlen(launcherdir) + strlen(filename) +1];
	sprintf (path, "%s/%s", launcherdir, filename);
	g_free (filename);
	
	char from_uri[strlen(uri)];
	sprintf (from_uri, "%s", uri);
	
	if (g_str_has_prefix (from_uri, file_prefix)) {
		gint i;
		gint diff;
		gint length;
		
		diff = strlen(file_prefix);
		length = strlen(from_uri);
		
		for (i = 0; i < length; i++) {
			if (i < length - diff) {
				from_uri[i] = from_uri[i+diff];
			} else {
				from_uri[i] = '\0';
				break;
			}
		}
	}
	
	if (g_file_test(path, G_FILE_TEST_EXISTS)) {
		if (!g_str_has_prefix(path, launcherdir)) {
			return NULL;
		} else {
			intern_dnd = TRUE;
		}
	}
	else {
		switch (action) {
			case GDK_ACTION_COPY: {
				if (!kiba_utils_copy_file(from_uri, path)) {
					return NULL;
				}
				
				break;
			}
			case GDK_ACTION_LINK: {
				link (from_uri, path);
				break;
			}
			case GDK_ACTION_MOVE: {
				g_rename (from_uri, path);
				break;
			}
			default:
				return;
				break;
		}
	}
	
	if (intern_dnd) {
		if (!group)
			return NULL;
		
		for (l = launcher->groups; l != NULL; l = l->next) {
			KibaLauncherIconGroup *check_group = l->data;
			GList *m;
			
			if (check_group == group) {
				continue;
			}
			
			for (m = check_group->icons; m != NULL; m = m->next) {
				KibaDesktopIcon *check_icon = KIBA_DESKTOP_ICON(m->data);
				
				if (strcmp (uri, check_icon->entry.uri) == 0) {
					kiba_launcher_remove (launcher, check_icon);
					g_object_unref (G_OBJECT(check_icon));
					intern_success = TRUE;
					break;
				}
			}
		}
	}
	if (intern_dnd && !intern_success) {
		return NULL;
	}
	
	if (group) {
		// remember the grop
		kiba_key_file_write_desktop_entry (path,
			KIBA_KEY_FILE_DESKTOP_KEY_GROUP, group->id);
	}
	
	// add a launcher icon
	icon = kiba_desktop_icon_new ();
	kiba_desktop_icon_set_file (icon, path);
	kiba_launcher_add (launcher, icon);
	// find icon group
	if (!group) {
		group = kiba_launcher_find_icon_group (launcher, icon);
	}
	
	// activate icon
	kiba_launcher_select_icon(launcher, group, icon);
	
	return icon;
}

gboolean
kiba_launcher_drag_receive_cb (KibaLauncher *launcher,
		KibaLauncherIconGroup *group,
		GdkDragContext *drag_context,
		const gchar *drag_data)
{
	KibaDesktopIcon *icon;
	gchar *string;
	
	gchar data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	// add icons for drag data
	while ((string = g_strrstr (data, "\n")) != NULL) {
		if (!g_str_has_suffix (string, ".desktop")) {
			string[0] = '\0';
			continue;
		}
		
		icon = kiba_launcher_add_icon_from_uri (launcher,
			group,
			drag_context->suggested_action,
			launcher->launcher_directory, string+1);
		
		string[0] = '\0';
	}
	
	if (g_str_has_suffix (data, ".desktop")) {
		icon = kiba_launcher_add_icon_from_uri (launcher,
			group,
			drag_context->suggested_action,
			launcher->launcher_directory, data);
	}
	
	return FALSE;
}

static void
kiba_launcher_drag_dest_func (GdkDragContext *drag_context,
		const gchar *drag_data,
		gpointer user_data,
		gchar **action_ret,
		gchar **icon_ret)
{
	gchar *string;
	gboolean desktop_file_found;
	
	gchar data[strlen(drag_data)];
	sprintf(data, "%s", drag_data);
	g_strstrip (data);
	
	desktop_file_found = FALSE;
	while ((string = g_strrstr (data, "\n")) != NULL) {
		if (!g_str_has_suffix (string, ".desktop")) {
			string[0] = '\0';
			continue;
		}
		
		desktop_file_found = TRUE;
		
		string[0] = '\0';
	}
	if (g_str_has_suffix (data, ".desktop")) {
		desktop_file_found = TRUE;
	}
	
	if (!desktop_file_found) {
		return;
	}
	
	*action_ret = g_strdup(_("Create Launcher"));
	*icon_ret = g_strdup(LAUNCHER_ICON);
}

static gboolean
kiba_launcher_drag_receive_func (GdkDragContext *drag_context,
		const gchar *drag_data,
		gpointer user_data)
{
	KibaLauncher *launcher = user_data;
	return kiba_launcher_drag_receive_cb (
		launcher, NULL, drag_context, drag_data);
}

static void
drag_received_cb (GtkWidget *widget,
		GdkDragContext *drag_context,
		gint x,
		gint y,
		GtkSelectionData *selection_data,
		guint info,
		guint time,
		gpointer user_data)
{
	KibaLauncher *launcher = user_data;
	KibaLauncherIconGroup *group;
	
	if (launcher->cfg_grouping) {
		group = find_object_group (launcher, KIBA_OBJECT(widget));
	} else {
		group = NULL;
	}
	
	kiba_launcher_drag_receive_cb (launcher,
		group,
		drag_context,
		(gchar *) selection_data->data);
	
	gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

/// Public Functions

KibaLauncherIconGroup*
kiba_launcher_find_icon_group (KibaLauncher *launcher,
		KibaDesktopIcon *desktop_icon)
{
	GList *l;
	
	g_return_val_if_fail(launcher != NULL, NULL);
	g_return_val_if_fail(KIBA_IS_DESKTOP_ICON(desktop_icon), NULL);
	
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		
		if (strcmp(group->id, launcher->cfg_grouping ?
		    desktop_icon->entry.group : desktop_icon->entry.name) == 0) {
			return group;
		}
	}
	
	return NULL;
}

void
kiba_launcher_select_icon(KibaLauncher *launcher,
		KibaLauncherIconGroup *group,
		KibaDesktopIcon *icon)
{
	kiba_object_set_title (group->object,
		icon->entry.name);
	
	if (group->icons->data == icon) {
		return;
	}

	while (group->icons->data != icon) {
		kiba_utils_glist_scroll (&group->icons, GDK_SCROLL_DOWN);
	}
	kiba_drawable_object_switch_surface(
		KIBA_DRAWABLE_OBJECT(group->object));
	kiba_icon_set_file(KIBA_ICON(group->object),
		icon->entry.icon);
	
	if (group->view_win) {
		icon_view_launcher_icon_selected_cb(launcher,
			icon, group);
	}
}

/**
 * kiba_launcher_ungroup_icons:
 * @launcher: the #KibaLauncher instance
 * 
 * Removes category groups, afterwards, each icon has on dock object.
 **/
void
kiba_launcher_ungroup_icons (KibaLauncher *launcher)
{
	GList *l;
	
	g_return_if_fail(launcher != NULL);
	
	// allow one icon per group
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		KibaDesktopIcon *group_icon;
		GList *li;
		
		if (KIBA_IS_ICON_VIEW_WIN(group->view_win)) {
			gtk_widget_destroy (group->view_win->win);
		}
		
		for (li = g_list_last(group->icons); li->prev != NULL; li = li->prev) {
			KibaDesktopIcon *icon = li->data;
			
			group->icons = g_list_remove (group->icons, li->data);
			group->count--;
		}
		
		group_icon = group->icons->data;
		g_free (group->id);
		group->id = g_strdup (group_icon->entry.name);
	}
	
	// create new groups
	for (l = launcher->icons; l != NULL; l = l->next) {
		KibaDesktopIcon *icon = l->data;
		
		if (kiba_launcher_find_icon_group (launcher, icon) == NULL) {
			create_icon_group (launcher, icon);
		}
	}
}

/**
 * kiba_launcher_group_icons:
 * @launcher: the #KibaLauncher instance
 * 
 * Groups launcher icons in categories.
 **/
void
kiba_launcher_group_icons (KibaLauncher *launcher)
{
	GList *l;
	GList *groups = NULL;
	
	g_return_if_fail(launcher != NULL);
	
	// create new group list
	for (l = launcher->groups; l != NULL; l = l->next) {
		KibaLauncherIconGroup *group = l->data;
		KibaDesktopIcon *group_icon;
		GList *m;
		
		group_icon = group->icons->data;
		
		for (m = groups; m != NULL; m = m->next) {
			KibaLauncherIconGroup *check_group = m->data;
			
			if (strcmp (check_group->id, group_icon->entry.group) == 0) {
				group->icons = g_list_remove (group->icons, l->data);
				g_list_free (group->icons);
				
				kiba_object_destroy (group->object);
				
				group->launcher = NULL;
				g_free (group->id);
				g_free (group);
				group = NULL;
				
				check_group->icons = g_list_append (check_group->icons, group_icon);
				check_group->count++;
				
				break;
			}
		}
		
		if (!group) {
			continue;
		}
		
		g_free (group->id);
		group->id = g_strdup (group_icon->entry.group);
		
		groups = g_list_append (groups, group);
	}
	
	// set new group list
	for (l = launcher->groups; l != NULL; l = launcher->groups) {
		launcher->groups = g_list_remove (launcher->groups, l->data);
	}
	g_list_free (launcher->groups);
	launcher->groups = groups;
}

/**
 * kiba_launcher_add:
 * @launcher: the #KibaLauncher instance
 * @icon: a #KibaDesktopIcon instance
 * 
 * adds a #KibaDesktopIcon to the launcher plugin.
 **/
void
kiba_launcher_add (KibaLauncher *launcher,
		KibaDesktopIcon *icon)
{
	KibaLauncherIconGroup *group;
	
	g_return_if_fail(launcher != NULL);
	g_return_if_fail(KIBA_IS_DESKTOP_ICON(icon));
	
	g_object_set_data(G_OBJECT(icon),
		ICON_LAUNCHER_KEY, launcher);
	
	launcher->icons = g_list_append (launcher->icons, icon);
	
	// connect to icon signals
	g_signal_connect (G_OBJECT(icon), "file-changed",
		G_CALLBACK(kiba_launcher_icon_file_set_cb), launcher);
	kiba_launcher_icon_file_set_cb (icon, icon->entry.uri, launcher);
	
	group = kiba_launcher_find_icon_group(launcher, icon);
	if (group && group->view_win) {
		icon_view_launcher_icon_added_cb(launcher,
			icon, group);
	}
}

/**
 * kiba_launcher_remove:
 * @launcher: the #KibaLauncher instance
 * @icon: a #KibaDesktopIcon instance
 * 
 * removes a previously added icon.
 **/
void
kiba_launcher_remove (KibaLauncher *launcher,
		KibaDesktopIcon *icon)
{
	KibaLauncherIconGroup *group;
	gboolean first;
	GList *l;
	
	g_return_if_fail(launcher != NULL);
	g_return_if_fail(KIBA_IS_DESKTOP_ICON(icon));
	
	// find group
	group = kiba_launcher_find_icon_group(launcher, icon);
	
	// disconnect from icon signals
	g_signal_handlers_disconnect_by_func (
		     G_OBJECT(icon),
		     G_CALLBACK(kiba_launcher_icon_file_set_cb),
		     launcher);
	
	if (group && group->view_win) {
		icon_view_launcher_icon_removed_cb(launcher,
				icon,
				group);
	}
	
	launcher->icons = g_list_remove (launcher->icons, icon);
	if (!group) {
		return;
	}
	first = (group->icons->data == icon);
	
	group->icons = g_list_remove (group->icons, icon);
	group->count--;
	
	if (!group->count) {
		gchar *tmp_id, *id = NULL;
		
		// the removed icon was the last one in the group
		
		// change the object id to allow adding an object with the same
		// id during destroying
		g_object_get(G_OBJECT(group->object),
			"object_id", &id, NULL);
		tmp_id = g_strdup_printf("%s_tmp", id);
		g_object_set(G_OBJECT(group->object),
			"object_id", tmp_id, NULL);
		g_free(id);
		g_free(tmp_id);
		
		kiba_object_destroy(group->object);
		
		g_list_free (group->icons);
		group->icons = NULL;
		
		launcher->groups = g_list_remove (launcher->groups, group);
		group->launcher = NULL;
		g_free (group->id);
	} else if (first) {
		// the removed icon was the first one in the group
		
		KibaDesktopIcon *new_active = group->icons->data;
		
		kiba_drawable_object_switch_surface (
			KIBA_DRAWABLE_OBJECT(group->object));
		kiba_icon_set_file(KIBA_ICON(group->object),
			new_active->entry.icon);
		kiba_object_set_title (group->object,
			new_active->entry.name);
	}
}

/// Class Functions

void
kiba_launcher_destroy (KibaLauncher *launcher)
{
	KibaWin *win = kiba_get_plugin_win(kiba_get(), launcher->plugin);
	KibaDock *dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	GList *l;
	
	for (l = launcher->icons; l != NULL; l = launcher->icons) {
		kiba_launcher_remove(launcher, KIBA_DESKTOP_ICON(l->data));
	}
	g_list_free (launcher->icons);
	g_list_free (launcher->groups);
	
	kiba_win_remove_drag_func (win, kiba_launcher_drag_receive_func);
	kiba_dock_remove_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock);
	kiba_dock_remove_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock_create);
	
	g_free(launcher->launcher_directory);
	
	g_free (launcher);
}

static void
kiba_launcher_init (KibaLauncher *launcher)
{
	launcher->groups = NULL;
	launcher->icons  = NULL;
	
	launcher->cfg_grouping = FALSE;
	launcher->cfg_scale = 1.0f;
	launcher->cfg_alpha = 1.0f;
}

KibaLauncher*
kiba_launcher_new (GTypeModule *type_module,
		const gchar* launcher_directory)
{
	KibaLauncher *launcher;
	KibaWin *win;
	KibaDock *dock;
	
	launcher = g_new0 (KibaLauncher, 1);
	kiba_launcher_init(launcher);
	
	launcher->plugin = KIBA_PLUGIN(type_module);
	launcher->launcher_directory = g_strdup(launcher_directory);
	
	win = kiba_get_plugin_win(kiba_get(), launcher->plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	// dock drag and drop handler
	kiba_win_add_drag_func (win,
		kiba_launcher_drag_receive_func,
		kiba_launcher_drag_dest_func,
		launcher);
	// dock menu entries
	kiba_dock_add_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock,
		LAUNCHER_ICON,
		_("Add Launcher"),
		launcher);
	kiba_dock_add_populate_func (dock,
		(KibaDockPopulateFunc) kiba_launcher_populate_dock_create,
		LAUNCHER_ICON,
		_("Create Launcher"),
		launcher);
	
	return launcher;
}
