/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-drawable-object.h>
#include <kiba-visual-feature.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#define _(string)   gettext (string)
#define PLUGIN_NAME "pulse"

#define OBJECT_SIZE(obj) MAX(GTK_WIDGET(obj)->allocation.width, GTK_WIDGET(obj)->allocation.height)

#define PULSE_RATE 40

typedef struct _KibaPulse KibaPulse;
typedef struct _KibaPulseObject KibaPulseObject;

struct _KibaPulseObject {
	KibaPulse *pulse;
	KibaObject *object;
	KibaVisualFeatureObject *vobject;
	gboolean entered;
	gboolean urgent;
	gfloat factor;
	gfloat last_factor;
	gint pulse_id;
};

typedef enum {
	KIBA_PULSE_URGENT = 0,
	KIBA_PULSE_HOVERING,
	KIBA_PULSE_TRIGGERING
}KibaPulseFlags;

struct _KibaPulse {
	KibaVisualFeature *feature;
	GList *objects;
	KibaPulseFlags flags;
	
	float cfg_factor;
	gint speed;
};

/// helper

static KibaPulseObject*
find_object (KibaPulse  *pulse,
		KibaObject *object) {
	GList *l;
	for (l = pulse->objects; l != NULL; l = l->next) {
		KibaPulseObject *pobject = l->data;
		if (pobject->object == object)
			return pobject;
	}
	return NULL;
}

static void
remove_object (KibaPulse *pulse,
		KibaObject *object)
{
	KibaPulseObject *pobject;
	
	kiba_visual_feature_remove_object(pulse->feature, object);
	
	pobject = find_object (pulse, object);
	if (pobject) {
		if (pobject->pulse_id) {
			g_source_remove (pobject->pulse_id);
		}
		pulse->objects = g_list_remove (pulse->objects, pobject);
		g_free (pobject);
	}
}

static KibaPulseObject*
add_object (KibaPulse *pulse,
		KibaObject *object)
{
	KibaPulseObject *pobject;
	
	pobject = g_new0 (KibaPulseObject, 1);
	pobject->pulse = pulse;
	pobject->object = object;
	pobject->factor = 1.0f;
	pobject->last_factor = 1.0f;
	pobject->entered = FALSE;
	pobject->urgent = FALSE;
	pobject->pulse_id = 0;
	pobject->vobject = kiba_visual_feature_add_object(pulse->feature, object);
	pulse->objects = g_list_append (pulse->objects, pobject);
	
	return pobject;
}

/// timeouts

static gint
pulse_object_timeout (gpointer data)
{
	KibaPulseObject *pobject = data;
	KibaPulse *pulse = pobject->pulse;
	
	pobject->factor = MIN(pulse->cfg_factor,
		pobject->factor + 2.0f*(0.02 + 0.02*pulse->speed/50.0));
	
	kiba_visual_feature_queue_resize(pulse->feature,
		pobject->vobject);
	gtk_widget_queue_resize_no_redraw(KIBA_WIN_CHILD(
		kiba_visual_feature_get_win(pulse->feature)));
	gtk_widget_queue_resize_no_redraw(GTK_WIDGET(pobject->object));
	
	if (pobject->factor == pulse->cfg_factor) {
		gboolean stop_pulse = FALSE;
		gboolean urgent;
		gboolean hovering;
		
		hovering = (GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING) && pobject->entered);
		if (!hovering) {
			urgent = (GET_FLAG(pulse->flags, KIBA_PULSE_URGENT) && pobject->urgent);
			if (!urgent) {
				stop_pulse = TRUE;
			}
		}
		
		pobject->factor = 1.0f;
		
		if (stop_pulse ||
		    GET_FLAG(pobject->object->state, KIBA_OBJECT_DRAGGING)) {
			// done with pulsing
			remove_object(pulse, pobject->object);
			pobject->pulse_id = 0;
			return FALSE;
		}
	}
	
	return TRUE;
}

static void
start_pulsing (KibaPulseObject *pobject)
{
	if (GET_FLAG(pobject->object->state, KIBA_OBJECT_DRAGGING)) {
		return;
	}
	
	if (!pobject->pulse_id) {
		pobject->pulse_id = g_timeout_add (PULSE_RATE, pulse_object_timeout, pobject);
	}
}

/// preferences callbacks

static void
hovering_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPulse *pulse = data;
	gboolean hovering = GPOINTER_TO_INT(value);
	
	if (hovering) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING)) {
			SET_FLAG(pulse->flags, KIBA_PULSE_HOVERING);
		}
	} else if (GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING)) {
		DEL_FLAG(pulse->flags, KIBA_PULSE_HOVERING);
	}
}

static void
triggering_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPulse *pulse = data;
	gboolean triggering = GPOINTER_TO_INT(value);
	
	if (triggering) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING)) {
			SET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING);
		}
	} else if (GET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING)) {
		DEL_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING);
	}
}

static void
urgent_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPulse *pulse = data;
	gboolean urgent = GPOINTER_TO_INT(value);
	
	if (urgent) {
		if (!GET_FLAG(pulse->flags, KIBA_PULSE_URGENT)) {
			SET_FLAG(pulse->flags, KIBA_PULSE_URGENT);
		}
	} else if (GET_FLAG(pulse->flags, KIBA_PULSE_URGENT)) {
		DEL_FLAG(pulse->flags, KIBA_PULSE_URGENT);
	}
}

static void
factor_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPulse *pulse = data;
	pulse->cfg_factor = g_ascii_strtod((char*) value, NULL);
}

static void
step_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaPulse *pulse = data;
	pulse->speed = GPOINTER_TO_INT(value);
}

/// widget event handler

static void
object_urgent_event_cb (KibaObject *object,
		gboolean urgent,
		gpointer data)
{
	KibaPulse *pulse = data;
	KibaPulseObject *pobject;
	
	pobject = find_object (pulse, object);
	if(urgent) {
		if (pobject) {
			pobject->urgent = TRUE;
		}
		if (GET_FLAG(pulse->flags, KIBA_PULSE_URGENT)) {
			if (!pobject) {
				pobject = add_object(pulse, object);
				pobject->urgent = TRUE;
			}
			start_pulsing(pobject);
		}
	} else {
		if (pobject) {
			pobject->urgent = FALSE;
		}
	}
}

static gboolean
object_release_event_cb (GtkWidget *widget,
		GdkEventButton *event,
		gpointer data)
{
	KibaPulse *pulse = data;
	KibaPulseObject *pobject;
	
	if (GET_FLAG(pulse->flags, KIBA_PULSE_TRIGGERING)) {
		pobject = find_object (pulse, KIBA_OBJECT(widget));
		if (!pobject) {
			pobject = add_object(pulse, KIBA_OBJECT(widget));
		}
		start_pulsing(pobject);
	}
	
	return FALSE;
}

static gboolean
object_enter_event_cb (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaPulse *pulse = data;
	KibaPulseObject *pobject;
	
	pobject = find_object (pulse, KIBA_OBJECT(widget));
	if (pobject) {
		pobject->entered = TRUE;
	}
	if (GET_FLAG(pulse->flags, KIBA_PULSE_HOVERING)) {
		if (!pobject) {
			pobject = add_object(pulse, KIBA_OBJECT(widget));
			pobject->entered = TRUE;
		}
		start_pulsing(pobject);
	}
	
	return FALSE;
}

static gboolean
object_leave_event_cb (GtkWidget  *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	KibaPulse *pulse = data;
	KibaPulseObject *pobject;
	
	pobject = find_object (pulse, KIBA_OBJECT(widget));
	if (pobject) {
		pobject->entered = FALSE;
	}
	
	return FALSE;
}

/// visual feature methods

static void
feature_object_added(
	KibaVisualFeature *feature,
	KibaObject *object)
{
	KibaPulse *pulse;
	
	if (!KIBA_IS_DRAWABLE_OBJECT(object)) {
		return;
	}
	
	pulse = g_object_get_data(G_OBJECT(feature), PLUGIN_NAME);
	
	// connect to object signals
	g_signal_connect (G_OBJECT(object), "enter-notify-event",
		G_CALLBACK(object_enter_event_cb), pulse);
	g_signal_connect (G_OBJECT(object), "leave-notify-event",
		G_CALLBACK(object_leave_event_cb), pulse);
	g_signal_connect (G_OBJECT(object), "button-release-event",
		G_CALLBACK(object_release_event_cb), pulse);
	g_signal_connect (G_OBJECT(object), "urgent-toggled",
		G_CALLBACK(object_urgent_event_cb), pulse);
}

static void
feature_object_removed(
		KibaVisualFeature *feature,
		KibaObject *object)
{
	KibaPulse *pulse;
	KibaPulseObject *pobject;
	
	pulse = g_object_get_data(G_OBJECT(feature), PLUGIN_NAME);
	
	// disconnect from object signals
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(object_enter_event_cb), pulse);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(object_leave_event_cb), pulse);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(object_release_event_cb), pulse);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(object_urgent_event_cb), pulse);
	
	pobject = find_object (pulse, object);
	if (pobject) {
		if (pobject->pulse_id) {
			g_source_remove (pobject->pulse_id);
		}
		pulse->objects = g_list_remove (pulse->objects, pobject);
		g_free (pobject);
	}
}

static void
feature_draw(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	cairo_surface_t *surface = drawable_object->rendered_object->surface;
	KibaPulseObject *pobject;
	KibaPulse *pulse;
	gfloat alpha;
	
	pulse = g_object_get_data(G_OBJECT(feature), PLUGIN_NAME);
	pobject = find_object (pulse, object->object);
	g_return_if_fail(pobject != NULL);
	
	alpha = kiba_drawable_object_get_alpha(drawable_object)*
	    (pulse->cfg_factor - pobject->last_factor)/(pulse->cfg_factor-1.0f);
	if (alpha <= 0.0f) {
		return;
	}
	
	// centered pulse
	cairo_scale(cr,
		pobject->last_factor,
		pobject->last_factor);
	
	// draw surface
	cairo_set_source_surface(cr, surface, 0, 0);
	cairo_paint_with_alpha(cr, alpha);
}

static KibaAllocation
feature_get_allocation(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	float surface_scale_x = 1.0f, surface_scale_y = 1.0f;
	KibaAllocation alloc;
	KibaPulseObject *pobject;
	KibaPulse *pulse;
	
	pulse = g_object_get_data(G_OBJECT(feature), PLUGIN_NAME);
	pobject = find_object (pulse, object->object);

	kiba_object_get_surface_scale(object->object,
		&surface_scale_x, &surface_scale_y);
	
	alloc = (KibaAllocation) {
		0.0f, 0.0f,
		object->object->alloc.width,
		object->object->alloc.height
	};
	
	if (pobject) {
		alloc.x += 0.5f*(1.0f-pobject->factor*surface_scale_x)*alloc.width;
		alloc.y += 0.5f*(1.0f-pobject->factor*surface_scale_y)*alloc.height;
		alloc.width *= pobject->factor * surface_scale_x;
		alloc.height *= pobject->factor * surface_scale_y;
		
		// remeber the last allocated scale factor,
		// the timeout that sets this var may be launched between size allocation
		// and redrawing.
		pobject->last_factor = pobject->factor;
	}
	
	return alloc;
}

static void
feature_get_max_offset(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		float *left, float *right,
		float *top, float *bottom)
{
	KibaDock *dock = kiba_visual_feature_get_dock(feature);
	float scale_x = 1.0f, scale_y = 1.0f;
	const KibaLayoutData *d;
	KibaPulseObject *pobject;
	KibaPulse *pulse;
	float dx, dy;
	GList *l;
	
	pulse = g_object_get_data(G_OBJECT(feature), PLUGIN_NAME);
	d = &kiba_layout_data[dock->d.position];
	pobject = find_object (pulse, object->object);
		
	for (l = object->object->scale_data; l != NULL; l = l->next) {
		KibaObjectScaleData *data = l->data;
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->max_func (object->object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		scale_x *= custom_factor_x;
		scale_y *= custom_factor_y;
	}
	scale_x *= object->object->cfg_scale_x;
	scale_y *= object->object->cfg_scale_y;
	
	if (pobject) {
		dx = 0.5f*(pulse->cfg_factor - 1.0f)*(object->object->size.width*scale_x);
		dy = 0.5f*(pulse->cfg_factor - 1.0f)*(object->object->size.height*scale_y);
	} else {
		dx = 0.0f;
		dy = 0.0f;
	}
	*left = dx;
	*right = dx;
	*bottom = dy;
	*top = dy;
}

/// Plugin Functions

static void plugin_unload (KibaPlugin *plugin)
{
	KibaPulse *pulse;
	
	pulse = g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	
	g_object_unref (pulse->feature);
	g_list_free (pulse->objects);
	g_free (pulse);
}

static gboolean
plugin_load (KibaPlugin *plugin)
{
	Kiba *kiba = kiba_get();
	KibaPrefsFile *file;
	KibaPulse *pulse;
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	pulse = g_new0(KibaPulse, 1);
	pulse->cfg_factor = 2.0;
	pulse->speed = 30;
	pulse->objects = NULL;
	
	// preferences callbacks
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "pulse_hovering",
		hovering_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_triggering",
		triggering_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_urgent",
		urgent_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_factor",
		factor_cb, pulse, TRUE);
	kiba_prefs_add_callback (file, "pulse_speed",
		step_cb, pulse, TRUE);
	
	pulse->feature = g_object_new (
		    KIBA_TYPE_VISUAL_FEATURE,
		    "name",  PLUGIN_NAME,
		    "type",  KIBA_VISUAL_FEATURE_GIMMICK,
		    "front", TRUE,
		    NULL);
	// KibaVisualFeature overrides
	pulse->feature->get_allocation = feature_get_allocation;
	pulse->feature->draw = feature_draw;
	pulse->feature->object_added = feature_object_added;
	pulse->feature->object_removed = feature_object_removed;
	pulse->feature->get_max_offset = feature_get_max_offset;
	// remember pulse on the feature
	g_object_set_data (G_OBJECT(pulse->feature),
		PLUGIN_NAME, (gpointer) pulse);
	// set feature win
	kiba_visual_feature_set_win(pulse->feature, win);
	
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) pulse);
	
	return TRUE;
}

static PrefMember pulse_options[] = {
    { "pulse_hovering", "TRUE", KIBA_BOOLEAN },
    { "pulse_triggering", "TRUE", KIBA_BOOLEAN },
    { "pulse_urgent", "TRUE", KIBA_BOOLEAN },
    { "pulse_factor", "1.5", KIBA_DOUBLE },
    { "pulse_speed", "30", KIBA_INT }
};

static KibaPluginVtable vtable = {
	"Pulse Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
	
	// plugin version
	MAJOR_VERSION,
	MINOR_VERSION
};

KIBA_INIT_PLUGIN (vtable, pulse_options, NULL, NULL, NULL);
