/*
 *  Copyright (C) 2007  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "config.h"

#include <libintl.h>
#include <math.h>
#include <gtk/gtk.h>

#include <kiba.h>
#include <kiba-win.h>
#include <kiba-dock.h>
#include <kiba-object.h>
#include <kiba-plugin.h>
#include <kiba-prefs.h>
#include <kiba-debug.h>

#include "switch.h"

#define PLUGIN_NAME "switch"

// localization
#define _(string) gettext (string)
// switch animation rate
#define SWITCH_RATE 20

/*** Helper Functions ***/

static void __get_object_features (KibaSwitchPlugin *switch_plugin,
		KibaSwitchPluginObject *switch_obj)
{
	GList *l;
	
	// find front feature object
	for (l = switch_plugin->front_feature->feature->objects; l != NULL; l = l->next) {
		KibaVisualFeatureObject *vobject = l->data;
		if (vobject->object == switch_obj->object) {
			switch_obj->front_feature_object = vobject;
			break;
		}
	}
	// find back feature object
	for (l = switch_plugin->back_feature->feature->objects; l != NULL; l = l->next) {
		KibaVisualFeatureObject *vobject = l->data;
		if (vobject->object == switch_obj->object) {
			switch_obj->back_feature_object = vobject;
			break;
		}
	}
}

static KibaSwitchPluginObject* __find_object (KibaSwitchPlugin *switch_plugin,
		KibaObject *object)
{
	GList *l;
	for (l = switch_plugin->objects; l != NULL; l = l->next) {
		KibaSwitchPluginObject *sobject = l->data;
		if (sobject->object == object)
			return sobject;
	}
	return NULL;
}

static void __switch_draw_surface (cairo_t *cr,
		cairo_surface_t  *surface,
		gint width,
		gint height,
		void *data)
{
	KibaSwitchPluginObject *switch_obj = data;
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(switch_obj->object);
	// using the current object render method for srface buffer
	drawable_object->render_object(drawable_object, cr);
}

static void __create_switch_surface(KibaSwitchPluginObject *switch_obj)
{
	cairo_t *cr;
	
	cr = kiba_win_cairo_create(switch_obj->switch_plugin->win);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return;
	
	// remember size
	switch_obj->switch_surface.width =
		switch_obj->object->size.width * switch_obj->object->cfg_scale_x;
	switch_obj->switch_surface.height =
		switch_obj->object->size.height * switch_obj->object->cfg_scale_y;
	// create surface buffer that will contain
	// the current displayed object unscaled
	cairo_update_surface_at_size (cr,
		&(switch_obj->switch_surface.surface),
		switch_obj->switch_surface.width,
		switch_obj->switch_surface.height,
		__switch_draw_surface, switch_obj);
}

static void __render_object(KibaDrawableObject *drawable_object, cairo_t *cr)
{
	KibaSwitchPluginObject *switch_obj =
		g_object_get_data(G_OBJECT(drawable_object), "switch-object");
	
	g_return_if_fail(switch_obj != NULL);
	
	// render the switched surface
	if (switch_obj->switch_surface.surface) {
		cairo_set_source_surface(cr,
			switch_obj->switch_surface.surface, 0, 0);
		cairo_paint(cr);
	}
}

static gint __switch_object_timeout (gpointer data)
{
	KibaSwitchPluginObject *object = data;
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(object->object);
	
	kiba_drawable_object_queue_draw (drawable_object);
	
	float min = object->switch_plugin->step/10.0f;
	object->last_step = (object->switch_plugin->step-min)*sin(
		(object->factor/2.0f + 0.5f)*M_PI) + min;
	
	// update scale factor
	if (object->switch_direction == 0) {
		object->factor -= object->last_step;
	} else {
		object->factor += object->last_step;
	}

	if (object->factor <= 0.05) {
		object->factor = 0.05;
		object->switch_direction = 1;
		// set saved render method again
		kiba_drawable_object_queue_render_object(drawable_object);
	    
		// update surface
		drawable_object->render_object = object->render_object_save;
		__create_switch_surface(object);
		drawable_object->render_object = __render_object;
	} else if (object->factor >= 1.0) {
		object->factor = 1.0;
		if (object->switch_queued) {
			object->switch_direction = 0;
			object->switch_queued = FALSE;
		} else {
			// reset render method
			drawable_object->render_object = object->render_object_save;
			object->render_object_save = NULL;
			
			kiba_object_update_scale_factor (object->object);
			kiba_drawable_object_queue_render_object(drawable_object);
			object->switch_id = 0;
			return FALSE;
		}
	}
	kiba_object_update_scale_factor (object->object);
	
	// queue draw for fake 3d objects
	if (object->switch_plugin->object3d) {
		kiba_visual_feature_queue_draw(
			object->switch_plugin->front_feature->feature,
			object->front_feature_object);
		kiba_visual_feature_queue_draw(
			object->switch_plugin->back_feature->feature,
			object->back_feature_object);
	}
	
	return TRUE;
}

static void __scale_func (KibaObject *object,
		float *scale_x,
		float *scale_y,
		gpointer data)
{
	KibaSwitchPluginObject *switch_object = data;
	
	// set scale factor depending on orientation
	if (GET_FLAG(switch_object->switch_plugin->orientation, GTK_ORIENTATION_VERTICAL)) {
		*scale_y = switch_object->factor;
		if (GET_FLAG(switch_object->switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL))
			*scale_x = switch_object->factor;
		else
			*scale_x = 1.0;
	} else {
		*scale_x = switch_object->factor;
		*scale_y = 1.0;
	}
}

static void __max_scale_func (KibaObject *object,
		float *scale_x,
		float *scale_y,
		gpointer data)
{
	// will never scale above 1.0f both directions
	*scale_x = 1.0;
	*scale_y = 1.0;
}

static void __switch_object (KibaDrawableObject *drawable_object,
		gpointer data)
{
	KibaSwitchPlugin *switch_plugin = data;
	KibaSwitchPluginObject *switch_obj;
	KibaObject *object = KIBA_OBJECT(drawable_object);
	
	switch_obj = __find_object (switch_plugin, object);

	if (!switch_obj) {
		switch_obj = g_new0(KibaSwitchPluginObject, 1);
		switch_obj->factor = 1.0;
		switch_obj->switch_plugin = switch_plugin;
		switch_obj->switch_id = 0;
		switch_obj->object = object;
		switch_obj->render_object_save = NULL;
		switch_obj->switch_surface.surface = NULL;
		switch_obj->switch_surface.width = 1;
		switch_obj->switch_surface.height = 1;
		switch_obj->front_feature_object = NULL;
		switch_obj->back_feature_object = NULL;
		
		switch_plugin->objects = g_list_append (switch_plugin->objects, switch_obj);
		
		g_object_set_data(G_OBJECT(drawable_object),
			"switch-object", switch_obj);
		
		kiba_object_add_scale_function (object,
			__scale_func,
			__max_scale_func,
			FALSE,
			switch_obj);
		
		if (switch_plugin->object3d)
			__get_object_features(switch_plugin, switch_obj);
	}
	
	if (switch_obj->render_object_save) {
		drawable_object->render_object = switch_obj->render_object_save;
	} else {
		__create_switch_surface(switch_obj);
	}
	
	// use custom surface for rendering
	switch_obj->render_object_save = drawable_object->render_object;
	drawable_object->render_object = __render_object;
	kiba_drawable_object_queue_render_object(drawable_object);
	
	if (!switch_obj->switch_id) {
		// remember current direction
		switch_obj->switch_direction = 0;
		switch_obj->switch_id = g_timeout_add (SWITCH_RATE,
			__switch_object_timeout, switch_obj);
	} else if (switch_obj->switch_direction == 1) {
		switch_obj->switch_queued = TRUE;
	}
}

/*** Signal Handler ***/

static void __prefs_3d_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaSwitchPlugin *switch_plugin = data;
	GList *l;
	
	switch_plugin->object3d = GPOINTER_TO_INT(value);
	
	// toggles fake 3d rendering
	if (!switch_plugin->object3d) {
		if (switch_plugin->front_feature)
			kiba_switch_3d_destroy(switch_plugin->front_feature);
		if (switch_plugin->back_feature)
			kiba_switch_3d_destroy(switch_plugin->back_feature);
		for (l = switch_plugin->objects; l != NULL; l = l->next) {
			KibaSwitchPluginObject *switch_obj = l->data;
			switch_obj->front_feature_object = NULL;
			switch_obj->back_feature_object = NULL;
		}
	} else {
		switch_plugin->front_feature = kiba_switch_3d_new(
			switch_plugin->win, switch_plugin, TRUE);
		switch_plugin->back_feature = kiba_switch_3d_new(
			switch_plugin->win, switch_plugin, FALSE);
		for (l = switch_plugin->objects; l != NULL; l = l->next) {
			KibaSwitchPluginObject *switch_obj = l->data;
			__get_object_features(switch_plugin, switch_obj);
		}
	}
}

static void __prefs_step_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaSwitchPlugin *switch_plugin = data;
	switch_plugin->step = g_ascii_strtod((char*) value, NULL);
}

static void __prefs_orientation_cb (const char *option_id,
		gpointer value,
		gpointer data)
{
	KibaSwitchPlugin *switch_plugin = data;
	char *orient = (char*) value;
	gboolean horizontal = FALSE, vertical = FALSE;
	
	if (strcmp (orient, "horizontal") == 0) {
		horizontal = TRUE;
	} else if (strcmp (orient, "vertical") == 0) {
		vertical = TRUE;
	} else if (strcmp (orient, "both") == 0) {
		vertical = TRUE;
		horizontal = TRUE;
	}
	
	if (horizontal) {
		if (!GET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL))
			SET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL);
	} else if (GET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL)) {
		DEL_FLAG(switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL);
	}
	if (vertical) {
		if (!GET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_VERTICAL))
			SET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_VERTICAL);
	} else if (GET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_VERTICAL)) {
		DEL_FLAG(switch_plugin->orientation, GTK_ORIENTATION_VERTICAL);
	}
}

static void __dock_object_add_cb (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	if (KIBA_IS_DRAWABLE_OBJECT(widget)) {
		g_signal_connect (G_OBJECT(widget), "switch-event",
			G_CALLBACK(__switch_object), data);
	}
}

static void __dock_object_remove_cb (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	KibaSwitchPlugin *switch_plugin = data;
	KibaSwitchPluginObject *object = __find_object (switch_plugin, KIBA_OBJECT(widget));
	
	if (!object) {
		return;
	}
	
	if (object->switch_id) {
		g_source_remove (object->switch_id);
		object->switch_id = 0;
	}
	
	if (object->factor != 1.0) {
		object->factor = 1.0;
		kiba_object_update_scale_factor (object->object);
	}
	
	kiba_object_remove_scale_function_by_data (object->object, object);
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(__switch_object), object->switch_plugin);
	
	object->switch_plugin->objects = g_list_remove (object->switch_plugin->objects, object);
	g_free (object);
}

/*** Plugin Functions ***/

static void plugin_unload (KibaPlugin *plugin)
{
	KibaSwitchPlugin *switch_plugin =
		g_object_get_data (G_OBJECT(plugin), PLUGIN_NAME);
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	
	for (l = switch_plugin->objects; l != NULL; l = switch_plugin->objects) {
		KibaSwitchPluginObject *object = l->data;
		switch_plugin->objects = g_list_remove (switch_plugin->objects, l->data);
		
		if (object->switch_id)
			g_source_remove (object->switch_id);
		
		if (object->factor != 1.0) {
			object->factor = 1.0;
			kiba_object_update_scale_factor (object->object);
		}
		
		if (object->switch_surface.surface != NULL) {
			cairo_surface_destroy(object->switch_surface.surface);
			object->switch_surface.surface = NULL;
		}
		
		kiba_object_remove_scale_function_by_data (object->object, object);
		
		g_free (object);
	}
	g_list_free (switch_plugin->objects);
	
	for (l = gtk_container_get_children(GTK_CONTAINER(dock)); l != NULL; l = l->next) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(l->data),
			G_CALLBACK(__switch_object), switch_plugin);
	}
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(__dock_object_add_cb), switch_plugin);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(__dock_object_remove_cb), switch_plugin);
	
	if (switch_plugin->object3d) {
		kiba_switch_3d_destroy(switch_plugin->front_feature);
		kiba_switch_3d_destroy(switch_plugin->back_feature);
	}
	
	g_free (switch_plugin);
}

static gboolean plugin_load (KibaPlugin *plugin)
{
	KibaSwitchPlugin *switch_plugin;
	Kiba *kiba = kiba_get();
	KibaWin *win;
	KibaDock *dock;
	GtkContainer *container;
	KibaPrefsFile *file;
	GList *l;
	
	win = kiba_get_plugin_win(kiba, plugin);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(win));
	container = GTK_CONTAINER(dock);
	
	// init custom datastructure
	switch_plugin = g_new0(KibaSwitchPlugin, 1);
	switch_plugin->win = win;
	switch_plugin->step = 0.5;
	switch_plugin->orientation = GTK_ORIENTATION_VERTICAL;
	switch_plugin->objects = NULL;
	switch_plugin->front_feature = NULL;
	switch_plugin->back_feature = NULL;
	switch_plugin->object3d = FALSE;
	
	// preferences callbacks
	file = kiba_get_plugin_config_file(kiba, plugin);
	kiba_prefs_add_callback (file, "switch_step",
		__prefs_step_cb, switch_plugin, TRUE);
	//kiba_prefs_add_callback (file, "switch_3d_objects",
	//	__prefs_3d_cb, switch_plugin, TRUE);
	kiba_prefs_add_callback (file, "switch_orientation",
		__prefs_orientation_cb, switch_plugin, TRUE);
	
	// callbacks on object add/remove
	g_signal_connect_after (G_OBJECT(dock), "add",
		G_CALLBACK(__dock_object_add_cb), switch_plugin);
	g_signal_connect_after (G_OBJECT(dock), "remove",
		G_CALLBACK(__dock_object_remove_cb), switch_plugin);
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		__dock_object_add_cb (container, child, switch_plugin);
	}
	
	// register the KibaSwitchPlugin type on the plugin
	g_object_set_data (G_OBJECT(plugin),
		PLUGIN_NAME, (gpointer) switch_plugin);
	
	return TRUE;
}

static PrefMember switch_options[] = {
    { "switch_step", "0.5", KIBA_DOUBLE },
    //{ "switch_3d_objects", "FALSE", KIBA_BOOLEAN },
    { "switch_orientation", "horizontal", KIBA_STRING },
};

static KibaPluginVtable vtable = {
	"Switch Plugin for Kiba-Dock", /* summary */
	NULL, /* description  */
	
	"Daniel Bessler <danielb@kiba-dock.org>", /* author */
	"http://www.kiba-dock.org", /* homepage */
	
	KIBA_PLUGIN_MULTIPLE_LOADABLE, /* loadable flag */
	
	plugin_load, /* load  */
	plugin_unload, /* unload */
	
	// plugin version
	MAJOR_VERSION,
	MINOR_VERSION
};

KIBA_INIT_PLUGIN (vtable, switch_options, NULL, NULL, NULL);
