
/*
 *  Copyright (C) 2009 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <daniel@orgizm.net>
*/

#include <cairo.h>
#include <math.h>

#include <kiba.h>
#include <kiba-drawable-object.h>
#include <kiba-prefs.h>
#include <kiba-dock.h>

#include "switch.h"

// must be even
#define OBJECT_DEPTH 12
// space between surfaces
#define SURFACE_SPACE 0.65f

static void __draw_3d_object(KibaSwitchPlugin *switch_plugin,
		KibaSwitchPluginObject *switch_obj,
		cairo_t *cr,
		float factor,
		float i)
{
	KibaDrawableObject *drawable_object = KIBA_DRAWABLE_OBJECT(switch_obj->object);
	float surface_scale_x = 1.0f, surface_scale_y = 1.0f;
	
	kiba_object_get_surface_scale(switch_obj->object, &surface_scale_x, &surface_scale_y);
	
	cairo_save(cr);
		cairo_scale(cr, 1.0f/surface_scale_x, 1.0f/surface_scale_y);
		if (GET_FLAG(switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL))
			cairo_translate(cr, i*factor*SURFACE_SPACE*(1.0f - switch_obj->factor), 0);
		else
			cairo_translate(cr, 0, i*factor*SURFACE_SPACE*(1.0f - switch_obj->factor));
		cairo_scale(cr, surface_scale_x, surface_scale_y);
		
		// paint object surface
		drawable_object->render_object (drawable_object, cr);
	cairo_restore (cr);
}

static void __draw_3d_back(KibaSwitchPlugin *switch_plugin,
		KibaSwitchPluginObject *switch_obj,
		cairo_t *cr)
{
	// factor for scale direction
	float factor;
	int i;
	
	if (switch_obj->switch_direction == 0) {
		factor = -1.0f;
	} else {
		factor = 1.0f;
	}
	
	for (i = OBJECT_DEPTH/2; i > 0; i--)
		__draw_3d_object(switch_plugin, switch_obj, cr, factor, (float) i);
}

static void __draw_3d_front(KibaSwitchPlugin *switch_plugin,
		KibaSwitchPluginObject *switch_obj,
		cairo_t *cr)
{
	// factor for scale direction
	float factor;
	int i;
	
	if (switch_obj->switch_direction == 0) {
		factor = 1.0f;
	} else {
		factor = -1.0f;
	}
	
	cairo_save(cr);
	for (i = 1; i < OBJECT_DEPTH/2 + 1; i++)
		__draw_3d_object(switch_plugin, switch_obj, cr, factor, (float) i);
	cairo_restore (cr);
}

static void
resize_feature_object(
		KibaVisualFeature *feature,
		KibaObject *object)
{
	
	GList *l;
	KibaVisualFeatureObject *vobject = NULL;
	for (l = feature->objects; l != NULL; l = l->next) {
		vobject = l->data;
		if (vobject->object == object) break;
	}
	if (vobject) {
		kiba_visual_feature_queue_resize(feature, vobject);
	}
}

static void
__scale_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

static void
__offset_changed_cb(KibaObject *object,
		gboolean widget_changed,
		gboolean surface_changed,
		gpointer data)
{
	resize_feature_object(data, object);
}

/*** KibaVisualFeature implementation */

static void kiba_switch_3d_feature_object_added(
		KibaVisualFeature *feature, KibaObject *object)
{
	if (!KIBA_IS_DRAWABLE_OBJECT(object))
		return;
	kiba_visual_feature_add_object(feature, object);
	g_signal_connect(G_OBJECT(object), "scale-changed",
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_connect(G_OBJECT(object), "offset-changed",
		G_CALLBACK(__offset_changed_cb), feature);
}

static void kiba_switch_3d_feature_object_removed(
		KibaVisualFeature *feature, KibaObject *object)
{
	kiba_visual_feature_remove_object(feature, object);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__scale_changed_cb), feature);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object),
		G_CALLBACK(__offset_changed_cb), feature);
}

static void kiba_switch_3d_feature_draw(KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		cairo_t *cr)
{
	KibaSwitchPlugin *switch_plugin;
	KibaSwitchPluginObject *switch_obj;
	KibaSwitch3d *switch_3d;
	GList *l;
	
	switch_3d = g_object_get_data(G_OBJECT(feature), "switch_3d");
	g_return_if_fail(switch_3d != NULL);
	
	switch_plugin = switch_3d->switch_plugin;
	
	switch_obj = g_object_get_data(G_OBJECT(object->object), "switch-object");
	if (switch_obj && switch_obj->factor < 1.0f) {
		float surface_scale_x = 1.0f, surface_scale_y = 1.0f;
		
		kiba_object_get_surface_scale(object->object, &surface_scale_x, &surface_scale_y);
		
		cairo_save(cr);
			cairo_scale(cr, 1.0f/surface_scale_x, 1.0f/surface_scale_y);
			cairo_translate(cr, -object->allocation.x, -object->allocation.y);
			cairo_scale(cr, surface_scale_x, surface_scale_y);
			
			if (switch_3d->front) {
				__draw_3d_front(switch_plugin, switch_obj, cr);
			} else {
				__draw_3d_back(switch_plugin, switch_obj, cr);
			}
		cairo_restore(cr);
	}
}

static KibaAllocation kiba_switch_3d_feature_get_allocation(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object)
{
	GdkRectangle area = kiba_drawable_object_get_area (
		KIBA_DRAWABLE_OBJECT(object->object));
	KibaSwitchPluginObject *switch_obj;
	float surface_scale_x = 1.0f, surface_scale_y = 1.0f;
	float widget_scale_x = 1.0f, widget_scale_y = 1.0f;
	
	kiba_object_get_surface_scale(object->object,
		&surface_scale_x, &surface_scale_y);
	kiba_object_get_widget_scale(object->object,
		&widget_scale_x, &widget_scale_y);
	
	switch_obj = g_object_get_data(G_OBJECT(object->object), "switch-object");
	if (switch_obj) {
		if (GET_FLAG(switch_obj->switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL))
			surface_scale_x /= switch_obj->factor;
		else
			surface_scale_y /= switch_obj->factor;
		object->allocation =(KibaAllocation) {
			0.0f, 0.0f,
			object->object->size.width*widget_scale_x*surface_scale_x,
			object->object->size.height*widget_scale_y*surface_scale_y
		};
		if (GET_FLAG(switch_obj->switch_plugin->orientation, GTK_ORIENTATION_HORIZONTAL))
			object->allocation.x -= (object->allocation.width-(float)area.width)/2.0f;
		else
			object->allocation.y -= (object->allocation.height-(float)area.height)/2.0f;
	} else {
		object->allocation =(KibaAllocation) {
			0.0f, 0.0f,
			area.width,
			area.height
		};
	}
	return object->allocation;
}

static void kiba_switch_3d_get_max_offset(
		KibaVisualFeature *feature,
		KibaVisualFeatureObject *object,
		float *left, float *right,
		float *top, float *bottom)
{
	*left = 0.0f;
	*right = 0.0f;
	*top = 0.0f;
	*bottom = 0.0f;
}


void kiba_switch_3d_destroy(KibaSwitch3d *switch_3d)
{
	if (switch_3d->front)
		kiba_win_remove_visual_feature(switch_3d->win,
			"switch_3d_front");
	else
		kiba_win_remove_visual_feature(switch_3d->win,
			"switch_3d_back");
	g_free(switch_3d);
}

KibaSwitch3d* kiba_switch_3d_new(
	KibaWin *win, KibaSwitchPlugin *switch_plugin,
	gboolean front)
{
	KibaVisualFeature *feature;
	KibaSwitch3d *switch_3d;
	
	switch_3d = g_new0(KibaSwitch3d, 1);
	switch_3d->win = win;
	switch_3d->switch_plugin = switch_plugin;
	switch_3d->front = front;
	if (front) {
		switch_3d->feature = g_object_new (
			    KIBA_TYPE_VISUAL_FEATURE,
			    "name",  "switch_3d_front",
			    "type",  KIBA_VISUAL_FEATURE_MORPH,
			    "front", TRUE,
			    NULL);
	} else {
		switch_3d->feature = g_object_new (
			    KIBA_TYPE_VISUAL_FEATURE,
			    "name",  "switch_3d_back",
			    "type",  KIBA_VISUAL_FEATURE_MORPH,
			    "front", FALSE,
			    NULL);
	}
	g_object_set_data(G_OBJECT(switch_3d->feature),
		"switch_3d", switch_3d);
	
	// KibaVisualFeature overrides
	switch_3d->feature->get_allocation = kiba_switch_3d_feature_get_allocation;
	switch_3d->feature->draw = kiba_switch_3d_feature_draw;
	switch_3d->feature->object_added = kiba_switch_3d_feature_object_added;
	switch_3d->feature->object_removed = kiba_switch_3d_feature_object_removed;
	switch_3d->feature->get_max_offset = kiba_switch_3d_get_max_offset;
	
	kiba_visual_feature_set_win(switch_3d->feature, win);
	
	return switch_3d;
}
