/*
 *  Copyright (C) 2007-2009  Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <math.h>

#include <kiba.h>
#include <kiba-title.h>
#include <kiba-info-win-manager.h>

#include "zoom.h"

#define SHRINK_RATE 40
#define GROW_RATE 20

typedef struct {
	KibaZoom *zoom;
	
	GList *objects;
	gboolean objects_added;
	
	// state flags
	gboolean entered;
	gboolean dock_pressed;
	gboolean ignore_dock_enter;
	gboolean ignore_scale_change;
	gboolean win_entered;
	
	// timeout ids
	gint grow_id;
	gint shrink_id;
	
	// last mouse position relative to root window
	gfloat mouse_x;
	gfloat mouse_y;
	
	gfloat magnitude;
}ParabolicZoom;

typedef struct {
	ParabolicZoom *pzoom;
	KibaObject *object;
	
	gfloat scale;
	gfloat last_scale;
	gfloat offset;
	
	gfloat scaled_position;
	gfloat new_position;
	gfloat new_size;
}ParabolicZoomObject;


static void
update (ParabolicZoom *pzoom);

/*** ***/

static ParabolicZoomObject* find_object (ParabolicZoom *pzoom,
		KibaObject *object)
{
	GList *l;
	for (l = pzoom->objects; l != NULL; l = l->next) {
		ParabolicZoomObject *zobject = l->data;
		if (zobject->object == object)
			return zobject;
	}
	return NULL;
}

static void
get_object_scale(KibaObject *object,
	float *scale_x, float *scale_y,
	gpointer data)
{
	ParabolicZoomObject *zobject = data;
	zobject->last_scale = zobject->scale;
	*scale_x = zobject->scale;
	*scale_y = zobject->scale;
}

static void
get_max_object_scale(KibaObject *object,
	float *max_scale_x, float *max_scale_y,
	gpointer data)
{
	ParabolicZoomObject *zobject = data;
	*max_scale_x = zobject->pzoom->zoom->cfg_magnitude;
	*max_scale_y = zobject->pzoom->zoom->cfg_magnitude; 
}

static void
get_object_offset(KibaObject *object,
	float *offset_x, float *offset_y,
	gpointer data)
{
	ParabolicZoomObject *zobject = data;
	KibaDock *dock = KIBA_WIN_DOCK(zobject->pzoom->zoom->win);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		*offset_x = zobject->offset;
		*offset_y = 0.0f;
	} else {
		*offset_x = 0.0f;
		*offset_y = zobject->offset;
	}
}

static void
get_max_object_offset(KibaObject *object,
	float *left, float *right,
	float *top, float *bottom,
	gpointer data)
{
	// dont need additional window space
	*top = 0.0f;
	*bottom = 0.0f;
	*left = 0.0f;
	*right = 0.0f;
}

static void
add_object(ParabolicZoom *pzoom,
		KibaObject *object)
{
	ParabolicZoomObject *zobject;
	
	zobject = g_new0(ParabolicZoomObject, 1);
	zobject->pzoom = pzoom;
	zobject->last_scale = 1.0f;
	zobject->object = object;
	zobject->scale = 1.0f;
	zobject->offset = 0.0f;
	zobject->scaled_position = 0.0f;
	zobject->new_position = 0.0f;
	
	kiba_object_add_offset_function (object,
		get_object_offset,
		get_max_object_offset,
		TRUE,
		zobject);
	kiba_object_add_scale_function (object,
		get_object_scale,
		get_max_object_scale,
		TRUE,
		zobject);
	
	pzoom->objects = g_list_append(pzoom->objects, zobject);
}

static void
add_objects(ParabolicZoom *pzoom)
{
	GList *l, *objects;
	KibaDock *dock = KIBA_WIN_DOCK(pzoom->zoom->win);
	
	objects = kiba_dock_get_objects(dock);
	
	pzoom->objects_added = TRUE;
	for (l = objects; l != NULL; l = l->next) {
		KibaObject *object = l->data;
		add_object(pzoom, object);
	}
}

static void
remove_object(ParabolicZoom *pzoom,
		KibaObject *object)
{
	GList *l;
	
	for (l = pzoom->objects; l != NULL; l = l->next) {
		ParabolicZoomObject *zobject = l->data;
		
		if (zobject->object != object) {
			continue;
		}
		
		pzoom->objects = g_list_remove(pzoom->objects, zobject);
		
		kiba_object_remove_scale_function_by_data (
			zobject->object, zobject);
		kiba_object_remove_offset_function_by_data (
			zobject->object, zobject);
		
		g_free(zobject);
		
		break;
	}
}

static void
remove_objects(ParabolicZoom *pzoom)
{
	GList *l;
	
	pzoom->objects_added = FALSE;
	for (l = pzoom->objects; l != NULL; l = pzoom->objects) {
		ParabolicZoomObject *zobject = l->data;
		pzoom->objects = g_list_remove(pzoom->objects, zobject);
		
		kiba_object_remove_scale_function_by_data (
			zobject->object, zobject);
		kiba_object_remove_offset_function_by_data (
			zobject->object, zobject);
		
		g_free(zobject);
	}
	g_list_free(pzoom->objects);
	pzoom->objects = NULL;
	// only start zooming if hovering an object
	pzoom->ignore_dock_enter = TRUE;
}

/*** parabolic scaling algorithm ***/

static gfloat
scaleIntegral (gfloat delta, gfloat fInfluence, gfloat fMagnitude)
{
	gfloat rate = fInfluence / M_PI_2;
	gfloat expansion = fMagnitude * rate;
	if (delta <= -fInfluence)
		return delta - expansion;
	if (delta >= +fInfluence)
		return delta + expansion;
	return delta + expansion * sin (delta / rate);
}

static gfloat
scale (gfloat delta, gfloat fInfluence, gfloat fMagnitude)
{
	gfloat rate = fInfluence / M_PI_2;
	if (delta <= -fInfluence)
		return 1.0;
	if (delta >= +fInfluence)
		return 1.0;
	return 1.0 + fMagnitude * cosf (delta / rate);
}

static gfloat
scaledPosition (gfloat fX, gfloat fMouseX, gfloat fInfluence, gfloat fMagnitude)
{
	return fMouseX + scaleIntegral (fX - fMouseX, fInfluence, fMagnitude);
}

static void
update (ParabolicZoom *pzoom)
{
	KibaDock *dock = KIBA_WIN_DOCK(pzoom->zoom->win);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	// last object position, start for new one
	gfloat last_position;
	gfloat last_zoomed_position;
	// mouse position relative to dock
	gfloat mouse_position;
	// unscaled space between objects
	gfloat gap;
	// start edge of dock
	gfloat start_position;
	// stop edge of dock
	gfloat end_position;
	// translation amount for each object
	gfloat adjustment;
	GList *l, *objects;
	gfloat mouse_x, mouse_y;
	
	objects = kiba_dock_get_objects(dock);
	gap = (float) dock->d.object_space;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		end_position = dock->shrinked_allocation.width;
	} else {
		end_position = dock->shrinked_allocation.height;
	}
	end_position += -dock->d.cap_size -gap;
	// pretending the start is at 0 (actually it is at dock->shrinked_allocation.x/y)
	start_position = dock->d.cap_size + gap;
	
	last_position = start_position;
	last_zoomed_position = last_position;
	
	mouse_x = pzoom->mouse_x -
		GTK_WIDGET(pzoom->zoom->win)->allocation.x;
	mouse_y = pzoom->mouse_y -
		GTK_WIDGET(pzoom->zoom->win)->allocation.y;
	
	// mouse postion depends on orientation
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		mouse_position = mouse_x
			// translate mouse position to the start of the dock
			// (there are maybe animations active that need additional window space)
			- dock->shrinked_allocation.x
			// for scrolled docks, translate against scroll amaount
			+ kiba_dock_get_scroll_offset(dock);
	} else {
		mouse_position = mouse_y
			// translate mouse position to the start of the dock
			// (there are maybe animations active that need additional window space)
			- dock->shrinked_allocation.y
			// for scrolled docks, translate against scroll amaount
			+ kiba_dock_get_scroll_offset(dock);
	}
	
	for (l = objects; l != NULL; l = l->next) {
		KibaObject *object = l->data;
		ParabolicZoomObject *zobject = find_object(pzoom, object);
		gfloat scale_x = 1.0f;
		gfloat scale_y = 1.0f;
		// position of object before update
		gfloat current_position;
		// size of object before update
		gfloat current_size;
		// new scale factor for objects
		gfloat new_scale;
		
		// current scale without parabolic zoom factor
		kiba_object_get_widget_scale(object, &scale_x, &scale_y);
		scale_x = object->cfg_scale_x*object->destroy_factor*object->widget_scale_x/zobject->last_scale;
		scale_y = object->cfg_scale_y*object->destroy_factor*object->widget_scale_y/zobject->last_scale;
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			current_size = (object->size.width * scale_x);
		} else {
			current_size = (object->size.height * scale_y);
		}
		
		// set start position to the last position
		float object_start_position = last_position;
		// set last position to the start of next object
		last_position += current_size + gap;
		// center of Object
		current_position = object_start_position + current_size*0.5f;
		// position of scaled object
		zobject->scaled_position = scaledPosition (
			current_position, mouse_position,
			pzoom->zoom->cfg_influence, pzoom->magnitude);
		// calculate/update scale fator
		new_scale = scale (current_position - mouse_position,
			pzoom->zoom->cfg_influence, pzoom->magnitude);
		if (new_scale != zobject->scale) {
			zobject->scale = new_scale;
			zobject->pzoom->ignore_scale_change = TRUE;
			kiba_object_update_scale_factor(zobject->object);
			zobject->pzoom->ignore_scale_change = FALSE;
		}
		zobject->new_size = current_size*new_scale;
		// calculate object position
		zobject->new_position = zobject->scaled_position -
			(current_position - object_start_position)*new_scale;
		// calculate offset to dock object allocation
		zobject->offset = zobject->new_position-last_zoomed_position;
		last_zoomed_position += zobject->new_size + gap;
	}
	
	// calculate adjustment amount
	if (g_list_length(objects) > 1) {
		ParabolicZoomObject *first;
		ParabolicZoomObject *last;
		float factor;
		
		first = find_object(pzoom,
			KIBA_OBJECT(objects->data));
		last = find_object(pzoom,
			KIBA_OBJECT(g_list_last(objects)->data));
		
		if (mouse_position < start_position) {
			// leftmost object should never be right of startposition
			adjustment = start_position - first->new_position;
		} else if (mouse_position > end_position - gap && last->new_position + last->new_size < mouse_position) {
			// rightmost object should never be left of endposition
			adjustment = end_position
				- last->new_position - last->new_size - gap*0.5f
				+ kiba_dock_get_scroll_offset(dock);
		} else {
			adjustment = 0.0f;
		}
	} else {
		adjustment = 0.0f;
	}
	
	// adjsut object positions
	for (l = objects; l != NULL; l = objects) {
		KibaObject *object = l->data;
		ParabolicZoomObject *zobject = find_object(pzoom, object);
		
		objects = g_list_remove(objects, object);
		
		zobject->offset += adjustment;
		kiba_object_update_offset(zobject->object);
	}
	g_list_free(objects);
}

/*** timeouts ***/

static gint
scale_down (gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (pzoom->entered && !pzoom->dock_pressed) {
		// mouse entered again
		pzoom->shrink_id = 0;
		return FALSE;
	}
	
	if (!pzoom->objects_added) {
		pzoom->magnitude = 0.0f;
		pzoom->shrink_id = 0;
		return FALSE;
	}
	
	// scale down magnitude
	if (pzoom->magnitude > 0.01f) {
		pzoom->magnitude *= 0.8f;
	} else {
		pzoom->magnitude = 0.0f;
	}
	
	// update offsets/scale factors
	update(pzoom);
	
	if (pzoom->magnitude == 0.0f) {
		// shrinked down
		pzoom->shrink_id = 0;
		// remove scale/offset method
		if (pzoom->objects_added) {
			remove_objects(pzoom);
		}
		return FALSE;
	}

	return TRUE;
}

static gint grow_up (gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	// grow up magnitude
	if (pzoom->magnitude < 0.01f) {
		pzoom->magnitude = 0.01f;
	} else if (pzoom->magnitude < pzoom->zoom->cfg_magnitude - 1.05f) {
		pzoom->magnitude /= 0.75f;
	} else {
		pzoom->magnitude = pzoom->zoom->cfg_magnitude - 1.0f;
	}
	
	// update offsets/scale factors
	update(pzoom);
	
	if (pzoom->magnitude == pzoom->zoom->cfg_magnitude - 1.0f) {
		// grown up
		pzoom->grow_id = 0;
		return FALSE;
	}

	return TRUE;
}

/*** event wrapper ***/

static void leave_event (ParabolicZoom *pzoom)
{
	if (pzoom->win_entered) {
		return;
	}
	
	// remember that we left
	pzoom->entered = FALSE;
	
	if (pzoom->magnitude == 0.0f) {
		// nothing there to shrink down
		return;
	}
	
	if (pzoom->grow_id) {
		g_source_remove (pzoom->grow_id);
		pzoom->grow_id = 0;
	}
	
	if (!pzoom->shrink_id) {
		pzoom->shrink_id = g_timeout_add (SHRINK_RATE, scale_down, pzoom);
	}
}

static void
enter_event (ParabolicZoom *pzoom)
{
	if (pzoom->dock_pressed) {
		return;
	}
	
	// add scale/offset method
	if (!pzoom->objects_added) {
		add_objects(pzoom);
	}
	
	// remember that we entered
	pzoom->entered = TRUE;
	
	if (pzoom->grow_id) {
		// allready growing
		return;
	}
	
	// remove shrink timeout
	if (pzoom->shrink_id) {
		g_source_remove (pzoom->shrink_id);
		pzoom->shrink_id = 0;
	}

	if (!pzoom->grow_id) {
		pzoom->grow_id = g_timeout_add (GROW_RATE, grow_up, pzoom);
	}
}

/*** gtk event handler ***/

static gboolean
object_leave_event (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	leave_event((ParabolicZoom*) data);
	return FALSE;
}

static gboolean
object_enter_event (GtkWidget *widget,
		GtkWidget *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	KibaDock *dock = KIBA_WIN_DOCK(pzoom->zoom->win);
	
	// FIXME: dont know why enter event occurs while dragging
	if (!kiba_dock_get_dragging_child(dock)) {
		pzoom->ignore_dock_enter = FALSE;
		enter_event(pzoom);
	}
	return FALSE;
}

static void
dock_add_event (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	g_signal_connect (G_OBJECT(widget),
		"enter-notify-event",
		G_CALLBACK(object_enter_event), pzoom);
	g_signal_connect (G_OBJECT(widget),
		"leave-notify-event",
		G_CALLBACK(object_leave_event), pzoom);
	
	if (pzoom->objects_added == TRUE) {
		add_object(pzoom, KIBA_OBJECT(widget));
	}
}

static void
dock_remove_event (GtkContainer *container,
		GtkWidget *widget,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_enter_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(widget),
		G_CALLBACK(object_leave_event), pzoom);
	
	if (pzoom->objects_added == TRUE) {
		remove_object(pzoom, KIBA_OBJECT(widget));
	}
}

static gboolean
dock_release_event (GtkWidget *dock,
		GdkEventButton *button,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	// dock is not pressed anymore
	pzoom->dock_pressed = FALSE;
	
	if (pzoom->entered) {
		// force enter event, for growing up
		enter_event(pzoom);
	}
	
	return FALSE;
}

static gboolean
dock_press_event (GtkWidget *dock,
		GdkEventButton *button,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (button->button == 3) {
		// dont care about context menus
		return FALSE;
	}
	
	// dock is pressed
	pzoom->dock_pressed = TRUE;
	pzoom->ignore_dock_enter = TRUE;
	
	if (pzoom->entered) {
		// force leave event for shrinking down
		leave_event(pzoom);
	}
	
	return FALSE;
}

static gboolean
event_on_dock(ParabolicZoom *pzoom,
		gint event_x,
		gint event_y)
{
	KibaDock *dock = KIBA_WIN_DOCK(pzoom->zoom->win);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	
	if (pzoom->win_entered) {
		return TRUE;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		if (dock->shrinked_allocation.x > event_x ||
		    dock->shrinked_allocation.x + dock->shrinked_allocation.width < event_x) {
			return FALSE;
		}
	} else {
		if (dock->shrinked_allocation.y > event_y ||
		    dock->shrinked_allocation.y + dock->shrinked_allocation.height < event_y) {
			return FALSE;
		}
	}
	return TRUE;
}

static gboolean
dock_leave_event (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	if (pzoom->entered) {
		leave_event(pzoom);
	}
	return FALSE;
}

static gboolean
dock_enter_event (GtkWidget  *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	pzoom->mouse_x = event->x_root;
	pzoom->mouse_y = event->y_root;
	
	if (event_on_dock(pzoom,
	    pzoom->mouse_x - GTK_WIDGET(pzoom->zoom->win)->allocation.x,
	    pzoom->mouse_y - GTK_WIDGET(pzoom->zoom->win)->allocation.y) &&
	   !pzoom->ignore_dock_enter) {
		enter_event(pzoom);
	}
	return FALSE;
}

static gboolean
dock_scroll_event (GtkWidget *dock,
		GdkEventScroll *scroll,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	pzoom->ignore_dock_enter = TRUE;
	
	if (pzoom->entered) {
		// force leave event for shrinking down
		leave_event(pzoom);
	}
	
	return FALSE;
}

static void
dock_size_changed(KibaDock *dock,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (!pzoom->objects_added) {
		return;
	}
	
	update(pzoom);
}

static gboolean
dock_motion_event (GtkWidget *widget,
		GdkEventMotion *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (!pzoom->objects_added && pzoom->ignore_dock_enter) {
		return TRUE;
	}
	
	pzoom->mouse_x = event->x_root;
	pzoom->mouse_y = event->y_root;
	
	if (event_on_dock(pzoom,
	    pzoom->mouse_x - GTK_WIDGET(pzoom->zoom->win)->allocation.x,
	    pzoom->mouse_y - GTK_WIDGET(pzoom->zoom->win)->allocation.y)) {
		// enter event if inside shrinked area and unadded
		if (!pzoom->objects_added && !pzoom->ignore_dock_enter) {
			enter_event(pzoom);
		}
	} else {
		// leave event if outside shrinked area and added
		if (pzoom->objects_added) {
			leave_event(pzoom);
		}
		return TRUE;
	}
	
	if (!pzoom->objects_added || pzoom->shrink_id) {
		return TRUE;
	}
	
	if (pzoom->magnitude != pzoom->zoom->cfg_magnitude-1.0f) {
		// objects still growing
		return TRUE;
	}

	update(pzoom);

	return TRUE;
}

static gboolean
title_leave_event (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	if (pzoom->win_entered) {
		return FALSE;
	}
	return dock_leave_event (
		KIBA_WIN_CHILD(pzoom->zoom->win),
		event, pzoom);
}

static gboolean
title_enter_event (GtkWidget  *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	if (pzoom->win_entered) {
		return FALSE;
	}
	return dock_enter_event(
		KIBA_WIN_CHILD(pzoom->zoom->win),
		event, pzoom);
}

static gboolean
title_motion_event (GtkWidget *widget,
		GdkEventMotion *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	return dock_motion_event (
		KIBA_WIN_CHILD(pzoom->zoom->win),
		event, pzoom);
}

static gboolean
info_win_leave_event (GtkWidget *widget,
		GdkEventCrossing *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	pzoom->win_entered = FALSE;
	return dock_leave_event (
		KIBA_WIN_CHILD(pzoom->zoom->win),
		event, pzoom);
}


static gboolean
info_win_motion_event (GtkWidget *widget,
		GdkEventMotion *event,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	pzoom->win_entered = TRUE;
	if (!pzoom->entered) {
		pzoom->mouse_x = event->x_root;
		pzoom->mouse_y = event->y_root;
		enter_event(pzoom);
	}
	
	return dock_motion_event (
		KIBA_WIN_CHILD(pzoom->zoom->win),
		event, pzoom);
}
		
static void
manager_win_add_event(
		KibaInfoWinManager *manager,
		KibaInfoWin *info_win,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (!info_win->attached) {
		return;
	}
	
	// connect to win signals
	g_signal_connect (G_OBJECT(info_win),
		"motion-notify-event",
		G_CALLBACK(info_win_motion_event), pzoom);
	g_signal_connect (G_OBJECT(info_win),
		"leave-notify-event",
		G_CALLBACK(info_win_leave_event), pzoom);
}

static void
manager_win_remove_event(
		KibaInfoWinManager *manager,
		KibaInfoWin *info_win,
		gpointer data)
{
	ParabolicZoom *pzoom = data;
	
	if (!info_win->attached) {
		return;
	}
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(info_win),
		G_CALLBACK(info_win_motion_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(info_win),
		G_CALLBACK(info_win_leave_event), pzoom);
	if (pzoom->win_entered) {
		info_win_leave_event(GTK_WIDGET(info_win), NULL, pzoom);
	}
}

/*** init/fini ***/

gpointer
parabolic_zoom_init(KibaZoom *zoom)
{
	ParabolicZoom *pzoom;
	KibaDock *dock = KIBA_WIN_DOCK(zoom->win);
	KibaTitle *title = kiba_get_title (kiba_get());
	KibaInfoWinManager *manager = kiba_info_win_manager_get();
	GList *l, *objects;
	
	objects = kiba_dock_get_objects(dock);
	
	pzoom = g_new0(ParabolicZoom, 1);
	pzoom->zoom = zoom;
	pzoom->magnitude = 0.0f;
	pzoom->shrink_id = 0;
	pzoom->grow_id = 0;
	pzoom->mouse_x = 0;
	pzoom->mouse_y = 0;
	pzoom->objects = NULL;
	pzoom->objects_added = FALSE;
	pzoom->entered = FALSE;
	pzoom->dock_pressed = FALSE;
	// only start zooming if hovering an object
	pzoom->ignore_dock_enter = TRUE;
	
	// connect to win manager signals
	g_signal_connect (G_OBJECT(manager), "info-win-removed",
		G_CALLBACK(manager_win_remove_event), pzoom);
	g_signal_connect (G_OBJECT(manager), "info-win-added",
		G_CALLBACK(manager_win_add_event), pzoom);
	// connect to title signals
	g_signal_connect (G_OBJECT(title),
		"motion-notify-event",
		G_CALLBACK(title_motion_event), pzoom);
	g_signal_connect (G_OBJECT(title),
		"enter-notify-event",
		G_CALLBACK(title_enter_event), pzoom);
	g_signal_connect (G_OBJECT(title),
		"leave-notify-event",
		G_CALLBACK(title_leave_event), pzoom);
	// connect to dock signals
	g_signal_connect (G_OBJECT(dock),
		"motion-notify-event",
		G_CALLBACK(dock_motion_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"enter-notify-event",
		G_CALLBACK(dock_enter_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"leave-notify-event",
		G_CALLBACK(dock_leave_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"button-press-event",
		G_CALLBACK(dock_press_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"button-release-event",
		G_CALLBACK(dock_release_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"scroll-event",
		G_CALLBACK(dock_scroll_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"size-changed",
		G_CALLBACK(dock_size_changed), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"add",
		G_CALLBACK(dock_add_event), pzoom);
	g_signal_connect (G_OBJECT(dock),
		"remove",
		G_CALLBACK(dock_remove_event), pzoom);
	for (l = objects; l != NULL; l = l->next) {
		dock_add_event(GTK_CONTAINER(dock),
			GTK_WIDGET(l->data), pzoom);
	}
	
	return pzoom;
}

void
parabolic_zoom_fini(gpointer data)
{
	ParabolicZoom *pzoom = data;
	KibaDock *dock = KIBA_WIN_DOCK(pzoom->zoom->win);
	KibaTitle *title = kiba_get_title (kiba_get());
	KibaInfoWinManager *manager = kiba_info_win_manager_get();
	GList *objects, *l;
	
	objects = kiba_dock_get_objects(dock);
	
	if (pzoom->shrink_id) {
		g_source_remove (pzoom->shrink_id);
		pzoom->shrink_id = 0;
	}
	if (pzoom->grow_id) {
		g_source_remove (pzoom->grow_id);
		pzoom->grow_id = 0;
	}
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(manager),
		G_CALLBACK(manager_win_remove_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(manager),
		G_CALLBACK(manager_win_add_event), pzoom);
	for (l = kiba_info_wins_get(); l != NULL; l = l->next) {
		KibaInfoWin *win = l->data;
		manager_win_remove_event(manager, win, pzoom);
	}
	g_signal_handlers_disconnect_by_func (G_OBJECT(title),
		G_CALLBACK(title_motion_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(title),
		G_CALLBACK(title_enter_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(title),
		G_CALLBACK(title_leave_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_motion_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_enter_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_leave_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_press_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_release_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_add_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_remove_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_scroll_event), pzoom);
	g_signal_handlers_disconnect_by_func (G_OBJECT(dock),
		G_CALLBACK(dock_size_changed), pzoom);
	
	for (l = objects; l != NULL; l = l->next) {
		dock_remove_event(GTK_CONTAINER(dock),
			GTK_WIDGET(l->data), pzoom);
	}
	
	g_free(pzoom);
}
